"""bigquery-etl CLI dag command."""

import os
import sys
from pathlib import Path

import rich_click as click
import yaml

from ..cli.utils import is_valid_dir, is_valid_file, sql_dir_option
from ..metadata.parse_metadata import METADATA_FILE, Metadata
from ..query_scheduling.dag import Dag
from ..query_scheduling.dag_collection import DagCollection
from ..query_scheduling.generate_airflow_dags import get_dags

dags_config_option = click.option(
    "--dags_config",
    "--dags-config",
    help="Path to dags.yaml config file",
    type=click.Path(file_okay=True),
    default="dags.yaml",
    callback=is_valid_file,
)

output_dir_option = click.option(
    "--output-dir",
    "--output_dir",
    help="Path directory with generated DAGs",
    type=click.Path(file_okay=False),
    default="dags/",
    callback=is_valid_dir,
)


@click.group(help="Commands for managing DAGs.")
def dag():
    """Create the CLI group for the dag command."""
    pass


@dag.command(
    help="""Get information about available DAGs.

    Examples:

    # Get information about all available DAGs
    ./bqetl dag info

    # Get information about a specific DAG
    ./bqetl dag info bqetl_ssl_ratios

    # Get information about a specific DAG including scheduled tasks
    ./bqetl dag info --with_tasks bqetl_ssl_ratios
    """,
)
@click.argument("name", required=False)
@dags_config_option
@sql_dir_option
@click.option(
    "--with_tasks",
    "--with-tasks",
    "-t",
    help="Include scheduled tasks",
    default=False,
    is_flag=True,
)
def info(name, dags_config, sql_dir, with_tasks):
    """List available DAG information."""
    if with_tasks:
        dag_collection = get_dags(None, dags_config, sql_dir=sql_dir)
    else:
        dag_collection = DagCollection.from_file(dags_config)

    if name:
        dag = dag_collection.dag_by_name(name)
        if not dag:
            click.echo(f"DAG {name} does not exist", err=True)
            sys.exit(1)
        sorted_dags = [dag]
    else:
        sorted_dags = sorted(dag_collection.dags, key=lambda d: d.name)

    for dag in sorted_dags:
        click.secho(dag.name, bold=True)
        click.echo(f"schedule_interval: {dag.schedule_interval}")
        click.echo(f"owner: {dag.default_args.owner}")

        if with_tasks:
            click.echo("tasks: ")
            for task in sorted(dag.tasks, key=lambda d: d.table):
                click.echo(
                    f"  - {task.project}.{task.dataset}.{task.table}_{task.version}"
                )

        click.echo("")


@dag.command(
    help="""Create a new DAG with name bqetl_<dag_name>, for example: bqetl_search
    When creating new DAGs, the DAG name must have a `bqetl_` prefix.
    Created DAGs are added to the `dags.yaml` file.

    Examples:

    \b
    ./bqetl dag create bqetl_core \\
    --schedule-interval="0 2 * * *" \\
    --owner=example@mozilla.com \\
    --description="Tables derived from `core` pings sent by mobile applications." \\
    --tag=impact/tier_1 \\
    --start-date=2019-07-25

    \b
    # Create DAG and overwrite default settings
    ./bqetl dag create bqetl_ssl_ratios --schedule-interval="0 2 * * *" \\
    --owner=example@mozilla.com \\
    --description="The DAG schedules SSL ratios queries." \\
    --tag=impact/tier_1 \\
    --start-date=2019-07-20 \\
    --email=example2@mozilla.com \\
    --email=example3@mozilla.com \\
    --retries=2 \\
    --retry_delay=30m
    """
)
@click.argument("name")
@dags_config_option
@click.option(
    "--schedule_interval",
    "--schedule-interval",
    help=(
        "Schedule interval of the new DAG. "
        "Schedule intervals can be either in CRON format or one of: "
        "once, hourly, daily, weekly, monthly, yearly or a timedelta []d[]h[]m"
    ),
    required=True,
)
@click.option(
    "--owner",
    help=("Email address of the DAG owner"),
    required=True,
)
@click.option(
    "--description",
    help=("Description for DAG"),
    required=True,
)
@click.option(
    "--tag",
    help=("Tag to apply to the DAG"),
    required=True,
    multiple=True,
)
@click.option(
    "--start_date",
    "--start-date",
    help=("First date for which scheduled queries should be executed"),
    required=True,
)
@click.option(
    "--email",
    help=("Email addresses that Airflow will send alerts to"),
    default=["telemetry-alerts@mozilla.com"],
    multiple=True,
)
@click.option(
    "--retries",
    help=("Number of retries Airflow will attempt in case of failures"),
    default=2,
)
@click.option(
    "--retry_delay",
    "--retry-delay",
    help=(
        "Time period Airflow will wait after failures before running failed tasks again"
    ),
    default="30m",
)
def create(
    name,
    dags_config,
    schedule_interval,
    owner,
    description,
    tag,
    start_date,
    email,
    retries,
    retry_delay,
):
    """Create a new DAG."""
    # create a DAG and validate all properties
    new_dag = Dag.from_dict(
        {
            name: {
                "description": description,
                "schedule_interval": schedule_interval,
                "default_args": {
                    "owner": owner,
                    "start_date": start_date,
                    "email": (*email, owner),
                    "retries": retries,
                    "retry_delay": retry_delay,
                },
                "tags": tag,
            }
        }
    )

    with open(dags_config, "a") as dags_file:
        dags_file.write("\n")
        dags_file.write(yaml.dump(new_dag.to_dict()))
        dags_file.write("\n")

    click.echo(f"Added new DAG definition to {dags_config}")


@dag.command(
    help="""Generate Airflow DAGs from DAG definitions.

    Examples:

    # Generate all DAGs
    ./bqetl dag generate

    # Generate a specific DAG
    ./bqetl dag generate bqetl_ssl_ratios
    """
)
@click.argument("name", required=False)
@dags_config_option
@sql_dir_option
@output_dir_option
def generate(name, dags_config, sql_dir, output_dir):
    """CLI command for generating Airflow DAGs."""
    dags = get_dags(None, dags_config, sql_dir)
    if name:
        # only generate specific DAG
        dag = dags.dag_by_name(name)

        if not dag:
            click.echo(f"DAG {name} does not exist.", err=True)
            sys.exit(1)

        dags.to_airflow_dags(output_dir, dag_to_generate=dag)
        click.echo(f"Generated {os.path.join(output_dir, dag.name)}.py")
    else:
        # re-generate all DAGs
        dags.to_airflow_dags(output_dir)
        click.echo("DAG generation complete.")


@dag.command(
    help="""Remove a DAG.
    This will also remove the scheduling information from the queries that were scheduled
    as part of the DAG.

    Examples:

    # Remove a specific DAG
    ./bqetl dag remove bqetl_vrbrowser
    """
)
@click.argument("name", required=False)
@dags_config_option
@sql_dir_option
@output_dir_option
def remove(name, dags_config, sql_dir, output_dir):
    """
    CLI command for removing a DAG.

    Also removes scheduling information from queries that were referring to the DAG.
    """
    # remove from task schedulings
    dags = get_dags(None, dags_config, sql_dir=sql_dir)
    dag_tbr = dags.dag_by_name(name)

    if not dag_tbr:
        click.echo(f"No existing DAG definition for {name}")
        sys.exit(1)

    for task in dag_tbr.tasks:
        metadata = Metadata.of_query_file(task.query_file)
        sql_path = Path(os.path.dirname(task.query_file))
        metadata.scheduling = {}
        metadata_file = sql_path / METADATA_FILE
        metadata.write(metadata_file)

    # remove from dags.yaml
    with open(dags_config) as dags_file:
        dags_config_dict = yaml.full_load(dags_file)
        del dags_config_dict[name]

        with open(dags_config, "w") as dags_file:
            dags_file.write(yaml.dump(dags_config_dict))

    output_dir = Path(output_dir)
    # delete generated DAG from dags/
    if os.path.exists(output_dir / (name + ".py")):
        os.remove(output_dir / (name + ".py"))

    click.echo(f"DAG {name} and referenced removed.")



"""bigquery-etl CLI metadata command."""

from datetime import datetime
from functools import partial
from multiprocessing.pool import Pool
from pathlib import Path
from typing import Optional

import click
from dateutil.relativedelta import relativedelta
from google.cloud import bigquery

from bigquery_etl.metadata.parse_metadata import DatasetMetadata, Metadata
from bigquery_etl.metadata.publish_metadata import publish_metadata

from ..cli.utils import (
    parallelism_option,
    paths_matching_name_pattern,
    project_id_option,
    sql_dir_option,
)
from ..config import ConfigLoader
from ..dryrun import get_credentials
from ..util import extract_from_query_path


@click.group(
    help="""
        Commands for managing bqetl metadata.
        """
)
@click.pass_context
def metadata(ctx):
    """Create the CLI group for the metadata command."""
    pass


@metadata.command(
    help="""
    Update metadata yaml files.
    Updates workgroup access metadata based on the dataset_metadata.yaml and
    deprecation metadata.

    Example:
     ./bqetl metadata update ga_derived.downloads_with_attribution_v2
    """,
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument("name")
@project_id_option()
@sql_dir_option
def update(name: str, sql_dir: Optional[str], project_id: Optional[str]) -> None:
    """Update metadata yaml file."""
    table_metadata_files = paths_matching_name_pattern(
        name, sql_dir, project_id=project_id, files=["metadata.yaml"]
    )
    retained_dataset_roles = ConfigLoader.get(
        "deprecation", "retain_dataset_roles", fallback=[]
    )

    # create and populate the dataset metadata yaml file if it does not exist
    for table_metadata_file in table_metadata_files:
        dataset_metadata_path = (
            Path(table_metadata_file).parent.parent / "dataset_metadata.yaml"
        )
        if not dataset_metadata_path.exists():
            continue
        dataset_metadata = DatasetMetadata.from_file(dataset_metadata_path)
        table_metadata = Metadata.from_file(table_metadata_file)

        dataset_metadata_updated = False
        table_metadata_updated = False

        # set dataset metadata default_table_workgroup_access to table_workgroup_access if not set
        if not dataset_metadata.default_table_workgroup_access:
            dataset_metadata.default_table_workgroup_access = (
                dataset_metadata.workgroup_access
            )
            dataset_metadata_updated = True

        if table_metadata.deprecated:
            # set workgroup: [] if table has been tagged as deprecated
            # this overwrites existing workgroups
            table_metadata.workgroup_access = []
            table_metadata_updated = True
            dataset_metadata.workgroup_access = [
                workgroup
                for workgroup in dataset_metadata.workgroup_access
                if workgroup.get("role") in retained_dataset_roles
            ]
            dataset_metadata_updated = True
        else:
            if table_metadata.workgroup_access is None:
                table_metadata.workgroup_access = (
                    dataset_metadata.default_table_workgroup_access
                )
                table_metadata_updated = True

        if dataset_metadata_updated:
            dataset_metadata.write(dataset_metadata_path)
            click.echo(f"Updated {dataset_metadata_path}")
        if table_metadata_updated:
            table_metadata.write(table_metadata_file)
            click.echo(f"Updated {table_metadata_file}")

    return None


@metadata.command(
    help="""
    Publish all metadata based on metadata.yaml file.

    Example:
     ./bqetl metadata publish ga_derived.downloads_with_attribution_v2
    """,
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument("name")
@project_id_option(
    ConfigLoader.get("default", "project", fallback="moz-fx-data-shared-prod")
)
@sql_dir_option
@parallelism_option()
def publish(
    name: str, sql_dir: Optional[str], project_id: Optional[str], parallelism: int
) -> None:
    """Publish Bigquery metadata."""
    table_metadata_files = paths_matching_name_pattern(
        name, sql_dir, project_id=project_id, files=["metadata.yaml"]
    )

    if parallelism > 0:
        credentials = get_credentials()

        with Pool(parallelism) as pool:
            pool.map(
                partial(_publish_metadata, project_id, credentials),
                table_metadata_files,
            )
    else:
        for metadata_file in table_metadata_files:
            _publish_metadata(project_id, credentials=None, metadata_file=metadata_file)


def _publish_metadata(project_id, credentials, metadata_file):
    project, dataset, table = extract_from_query_path(metadata_file)
    try:
        metadata = Metadata.from_file(metadata_file)
        publish_metadata(
            bigquery.Client(project=project_id, credentials=credentials),
            project,
            dataset,
            table,
            metadata,
        )
    except FileNotFoundError:
        print("No metadata file for: {}.{}.{}".format(project, dataset, table))


@metadata.command(
    help="""
    Deprecate BigQuery table by updating metadata.yaml file.
    Deletion date is by default 3 months from current date if not provided.

    Example:
     ./bqetl metadata deprecate ga_derived.downloads_with_attribution_v2 --deletion_date=2024-03-02
    """
)
@click.argument("name")
@project_id_option(
    ConfigLoader.get("default", "project", fallback="moz-fx-data-shared-prod")
)
@sql_dir_option
@click.option(
    "--deletion_date",
    "--deletion-date",
    help="Date when table is scheduled for deletion. Date format: yyyy-mm-dd",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    default=datetime.today() + relativedelta(months=+3),
)
def deprecate(
    name: str,
    sql_dir: str,
    project_id: str,
    deletion_date: datetime,
) -> None:
    """Deprecate Bigquery table by updating metadata yaml file(s)."""
    table_metadata_files = paths_matching_name_pattern(
        name, sql_dir, project_id=project_id, files=["metadata.yaml"]
    )

    for metadata_file in table_metadata_files:
        metadata = Metadata.from_file(metadata_file)

        metadata.deprecated = True
        metadata.deletion_date = deletion_date.date()

        metadata.write(metadata_file)
        click.echo(f"Updated {metadata_file} with deprecation.")

    if not table_metadata_files:
        raise FileNotFoundError(f"No metadata file(s) were found for: {name}")



from pathlib import Path

import pytest

from bigquery_etl.metadata.parse_metadata import Metadata
from bigquery_etl.query_scheduling.dag import (
    Dag,
    DagDefaultArgs,
    DagParseException,
    PublicDataJsonDag,
)
from bigquery_etl.query_scheduling.dag_collection import DagCollection
from bigquery_etl.query_scheduling.task import Task

TEST_DIR = Path(__file__).parent.parent


class TestDag:
    default_args = {
        "owner": "test@example.org",
        "email": ["test@example.org"],
        "depends_on_past": False,
    }

    def test_dag_instantiation(self):
        dag = Dag("bqetl_test_dag", "daily", self.default_args)

        assert dag.name == "bqetl_test_dag"
        assert dag.schedule_interval == "daily"
        assert dag.tasks == []
        assert dag.default_args == self.default_args

    def test_add_tasks(self):
        dag = Dag("bqetl_test_dag", "daily", self.default_args)

        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        scheduling = {
            "dag_name": "bqetl_test_dag",
            "default_args": {"owner": "test@example.org"},
            "task_name": "custom_task_name",
        }

        metadata = Metadata("test", "test", ["test@example.org"], {}, scheduling)

        task1 = Task.of_query(query_file)
        task2 = Task.of_query(query_file, metadata)

        assert dag.tasks == []

        dag.add_tasks([task1, task2])

        assert len(dag.tasks) == 2

    def test_add_tasks_duplicate_names(self):
        dag = Dag("bqetl_test_dag", "daily", self.default_args)

        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        scheduling = {
            "dag_name": "bqetl_test_dag",
            "default_args": {"owner": "test@example.org"},
            "task_name": "duplicate_task_name",
        }

        metadata = Metadata("test", "test", ["test@example.org"], {}, scheduling)
        task = Task.of_query(query_file, metadata)

        with pytest.raises(ValueError):
            dag.add_tasks([task, task])

    def test_from_dict(self):
        dag = Dag.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "start_date": "2020-05-12",
                        "owner": "test@example.com",
                        "email": ["test@example.com"],
                    },
                }
            }
        )

        assert dag.name == "bqetl_test_dag"
        assert dag.schedule_interval == "daily"
        assert dag.default_args.owner == "test@example.com"
        assert dag.default_args.email == ["test@example.com"]
        assert dag.default_args.depends_on_past is False

    def test_from_empty_dict(self):
        with pytest.raises(DagParseException):
            Dag.from_dict({})

    def test_from_dict_multiple_dags(self):
        with pytest.raises(DagParseException):
            Dag.from_dict({"bqetl_test_dag1": {}, "bqetl_test_dag2": {}})

    def test_from_dict_without_scheduling_interval(self):
        with pytest.raises(DagParseException):
            Dag.from_dict({"bqetl_test_dag": {}})

    def test_invalid_dag_name(self):
        with pytest.raises(ValueError):
            Dag("test_dag", "daily", self.default_args)

    def test_private_dag_name(self):
        dag = Dag("private_bqetl_test_dag", "daily", self.default_args)

        assert dag.name == "private_bqetl_test_dag"

    def test_schedule_interval_format(self):
        assert Dag("bqetl_test_dag", "daily", self.default_args)
        assert Dag("bqetl_test_dag", "weekly", self.default_args)
        assert Dag("bqetl_test_dag", "once", self.default_args)

        with pytest.raises(ValueError):
            assert Dag("bqetl_test_dag", "never", self.default_args)

        assert Dag("bqetl_test_dag", "* * * * *", self.default_args)
        assert Dag("bqetl_test_dag", "1 * * * *", self.default_args)

        with pytest.raises(TypeError):
            assert Dag("bqetl_test_dag", 323, self.default_args)

    def test_empty_dag_default_args(self):
        with pytest.raises(TypeError):
            assert DagDefaultArgs()

    def test_dag_default_args_owner_validation(self):
        with pytest.raises(ValueError):
            assert DagDefaultArgs(owner="invalid_email", start_date="2020-12-12")

        with pytest.raises(ValueError):
            assert DagDefaultArgs(owner="test@example-com.", start_date="2020-12-12")

        with pytest.raises(TypeError):
            assert DagDefaultArgs(owner=None, start_date="2020-12-12")

        assert DagDefaultArgs(owner="test@example.org", start_date="2020-12-12")

    def test_dag_default_args_email_validation(self):
        with pytest.raises(ValueError):
            assert DagDefaultArgs(
                owner="test@example.com",
                start_date="2020-12-12",
                email=["valid@example.com", "invalid", "valid2@example.com"],
            )

        assert DagDefaultArgs(
            owner="test@example.com", start_date="2020-12-12", email=[]
        )

        assert DagDefaultArgs(
            owner="test@example.com",
            start_date="2020-12-12",
            email=["test@example.org", "test2@example.org"],
        )

    def test_dag_default_args_start_date_validation(self):
        with pytest.raises(ValueError):
            assert DagDefaultArgs(
                owner="test@example.com", start_date="March 12th 2020"
            )

        with pytest.raises(ValueError):
            assert DagDefaultArgs(owner="test@example.com", start_date="foo")

        with pytest.raises(ValueError):
            assert DagDefaultArgs(owner="test@example.com", start_date="2020-13-01")

        assert DagDefaultArgs(owner="test@example.com", start_date="2020-12-12")

    def test_dag_default_args_retry_delay_validation(self):
        with pytest.raises(ValueError):
            assert DagDefaultArgs(
                owner="test@example.com", start_date="2020-12-12", retry_delay="90"
            )

        with pytest.raises(ValueError):
            assert DagDefaultArgs(
                owner="test@example.com", start_date="2020-12-12", retry_delay="1d1h1m"
            )

        assert DagDefaultArgs(
            owner="test@example.com", start_date="2020-12-12", retry_delay="3h15m1s"
        )

        assert DagDefaultArgs(
            owner="test@example.com", start_date="2020-12-12", retry_delay="3h"
        )

    def test_public_data_json_dag_add_task(self):
        public_json_dag = PublicDataJsonDag(
            "bqetl_public_data_json_dag", "daily", self.default_args
        )

        assert len(public_json_dag.tasks) == 0

        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        task = Task.of_query(query_file)
        dag = Dag(
            "bqetl_events",
            "0 1 * * *",
            DagDefaultArgs("test@example.org", "2020-01-01"),
            [task],
        )
        public_json_dag.add_export_tasks([task], DagCollection([dag]))

        assert len(public_json_dag.tasks) == 1
        assert (
            public_json_dag.tasks[0].task_name
            == "export_public_data_json_test__incremental_query__v1"
        )
        assert public_json_dag.tasks[0].dag_name == "bqetl_public_data_json_dag"
        assert len(public_json_dag.tasks[0].dependencies) == 1



"""Testing data for query."""

from pathlib import Path

import yaml

ROOT = Path(__file__).parent

UUID = "df735f02-efe5-4b07-b212-583bb99ba241"
SUBMISSION_DATE = "2020-10-01"
APP_BUILD_ID = "2020100100"

LATEST_VERSION = [dict(channel="nightly", latest_version=84)]

# NOTE: what happens when channel = "*"?
CLIENTS_SCALAR_AGGREGATES = [
    {
        "client_id": UUID,
        "ping_type": "metrics",
        "os": "Android",
        "app_version": 84,
        "app_build_id": APP_BUILD_ID,
        "channel": "nightly",
        "scalar_aggregates": [],
    }
]

CLIENTS_DAILY_SCALAR_AGGREGATES = [
    {
        "submission_date": SUBMISSION_DATE,
        "client_id": UUID,
        "ping_type": "metrics",
        "os": "Android",
        "app_version": 84,
        "app_build_id": APP_BUILD_ID,
        "channel": "nightly",
        "scalar_aggregates": [
            {
                "metric": "metrics_has_recent_pwas",
                "metric_type": "boolean",
                "key": "",
                "agg_type": "false",
                "value": 1.0,
            },
            {
                "metric": "places_manager_write_query_count",
                "metric_type": "counter",
                "key": "",
                "agg_type": "count",
                "value": 3.0,
            },
            {
                "metric": "places_manager_write_query_count",
                "metric_type": "counter",
                "key": "",
                "agg_type": "sum",
                "value": 48.0,
            },
        ],
    }
]

EXPECT = [
    {
        "client_id": UUID,
        "ping_type": "metrics",
        "os": "Android",
        "app_version": 84,
        "app_build_id": APP_BUILD_ID,
        "channel": "nightly",
        "scalar_aggregates": [
            {
                "metric": "metrics_has_recent_pwas",
                "metric_type": "boolean",
                "key": "",
                "agg_type": "false",
                "value": 1.0,
            },
            {
                "metric": "places_manager_write_query_count",
                "metric_type": "counter",
                "key": "",
                "agg_type": "count",
                "value": 3.0,
            },
            {
                "metric": "places_manager_write_query_count",
                "metric_type": "counter",
                "key": "",
                "agg_type": "sum",
                "value": 48.0,
            },
            # extra field from averages
            {
                "metric": "places_manager_write_query_count",
                "metric_type": "counter",
                "key": "",
                "agg_type": "avg",
                "value": 16.0,
            },
        ],
    }
]

prefix = "glam_etl"
tables = [
    (
        f"{prefix}.org_mozilla_fenix_glam_nightly__latest_versions_v1.yaml",
        LATEST_VERSION,
    ),
    (
        f"{prefix}.org_mozilla_fenix_glam_nightly__clients_scalar_aggregates_v1.yaml",
        CLIENTS_SCALAR_AGGREGATES,
    ),
    (
        f"{prefix}.org_mozilla_fenix_glam_nightly__"
        "view_clients_daily_scalar_aggregates_v1.yaml",
        CLIENTS_DAILY_SCALAR_AGGREGATES,
    ),
    ("expect.yaml", EXPECT),
]
for name, data in tables:
    with (ROOT / name).open("w") as fp:
        yaml.dump(data, fp)



"""Utility functions used by the CLI."""

import fnmatch
import os
import re
from fnmatch import fnmatchcase
from glob import glob
from pathlib import Path
from typing import Iterator, List, Optional, Tuple

import click
from google.auth.exceptions import DefaultCredentialsError
from google.cloud import bigquery

from bigquery_etl.config import ConfigLoader
from bigquery_etl.util.common import TempDatasetReference, project_dirs

QUERY_FILE_RE = re.compile(
    r"^.*/([a-zA-Z0-9-]+)/([a-zA-Z0-9_]+)/([a-zA-Z0-9_]+(_v[0-9]+)?)/"
    r"(?:query\.sql|shredder_mitigation_query\.sql|part1\.sql|script\.sql|query\.py|view\.sql|metadata\.yaml|backfill\.yaml)$"
)
CHECKS_FILE_RE = re.compile(
    r"^.*/([a-zA-Z0-9-]+)/([a-zA-Z0-9_]+)/([a-zA-Z0-9_]+(_v[0-9]+)?)/"
    r"(?:checks\.sql)$"
)


def is_valid_dir(ctx, param, value):
    """Check if the parameter provided via click is an existing directory."""
    if not os.path.isdir(value) or not os.path.exists(value):
        raise click.BadParameter(f"Invalid directory path to {value}")
    return value


def is_valid_file(ctx, param, value):
    """Check if the parameter provided via click is an existing file."""
    if not os.path.isfile(value) or not os.path.exists(value):
        raise click.BadParameter(f"Invalid file path to {value}")
    return value


def is_authenticated():
    """Check if the user is authenticated to GCP."""
    try:
        bigquery.Client(project="")
    except DefaultCredentialsError:
        return False
    return True


def is_valid_project(ctx, param, value):
    """Check if the provided project_id corresponds to an existing project."""
    if (
        value is None
        or value
        in [Path(p).name for p in project_dirs()]
        + [
            ConfigLoader.get("default", "test_project"),
            ConfigLoader.get("default", "user_facing_project", fallback="mozdata"),
        ]
        + ConfigLoader.get("default", "additional_projects", fallback=[])
        or value.startswith(ConfigLoader.get("default", "backfill_project"))
    ):
        return value
    raise click.BadParameter(f"Invalid project {value}")


def table_matches_patterns(pattern, invert, table):
    """Check if tables match pattern."""
    if not isinstance(pattern, list):
        pattern = [pattern]

    matching = False
    for p in pattern:
        compiled_pattern = re.compile(fnmatch.translate(p))
        if compiled_pattern.match(table) is not None:
            matching = True
            break

    return matching != invert


def paths_matching_checks_pattern(
    pattern: str, sql_path: Optional[str], project_id: Optional[str]
) -> Iterator[Tuple[Path, str, str, str]]:
    """Return single path to checks.sql matching the name pattern."""
    checks_files = paths_matching_name_pattern(
        pattern, sql_path, project_id, ["checks.sql"], CHECKS_FILE_RE
    )

    if checks_files:
        for checks_file in checks_files:
            match = CHECKS_FILE_RE.match(str(checks_file))
            if match:
                project = match.group(1)
                dataset = match.group(2)
                table = match.group(3)
            yield checks_file, project, dataset, table
    else:
        print(f"No checks.sql file found in {sql_path}/{project_id}/{pattern}")


def paths_matching_name_pattern(
    pattern, sql_path, project_id, files=["*.sql"], file_regex=QUERY_FILE_RE
) -> List[Path]:
    """Return paths to queries matching the name pattern."""
    matching_files: List[Path] = []

    if pattern is None:
        pattern = "*.*"

    # click nargs are passed in as a tuple
    if isinstance(pattern, tuple) or isinstance(pattern, list):
        for p in pattern:
            matching_files += paths_matching_name_pattern(
                str(p), sql_path, project_id, files, file_regex
            )
    elif os.path.isdir(pattern):
        for root, _, _ in os.walk(pattern, followlinks=True):
            for file in files:
                matching_files.extend(
                    map(Path, glob(f"{root}/**/{file}", recursive=True))
                )
    elif os.path.isfile(pattern):
        matching_files.append(Path(pattern))
    else:
        sql_path = Path(sql_path)
        if project_id is not None:
            sql_path = sql_path / project_id

        all_matching_files: List[Path] = []

        for file in files:
            all_matching_files.extend(
                map(Path, glob(f"{sql_path}/**/{file}", recursive=True))
            )

        for query_file in all_matching_files:
            match = file_regex.match(str(query_file))
            if match:
                project = match.group(1)
                dataset = match.group(2)
                table = match.group(3)
                query_name = f"{project}.{dataset}.{table}"
                if fnmatchcase(query_name, f"*{pattern}"):
                    matching_files.append(query_file)
                elif project_id and fnmatchcase(query_name, f"{project_id}.{pattern}"):
                    matching_files.append(query_file)

    if len(matching_files) == 0:
        print(f"No files matching: {pattern}, {files}")

    return matching_files


sql_dir_option = click.option(
    "--sql_dir",
    "--sql-dir",
    help="Path to directory which contains queries.",
    type=click.Path(file_okay=False),
    default=ConfigLoader.get("default", "sql_dir", fallback="sql"),
    callback=is_valid_dir,
)


use_cloud_function_option = click.option(
    "--use_cloud_function",
    "--use-cloud-function",
    help=(
        "Use the Cloud Function for dry running SQL, if set to `True`. "
        "The Cloud Function can only access tables in shared-prod. "
        "If set to `False`, use active GCP credentials for the dry run."
    ),
    type=bool,
    default=True,
)


def parallelism_option(default=8):
    """Generate a parallelism option, with optional default."""
    return click.option(
        "--parallelism",
        "-p",
        default=default,
        type=int,
        help="Number of threads for parallel processing",
    )


def project_id_option(default=None, required=False):
    """Generate a project-id option, with optional default."""
    return click.option(
        "--project-id",
        "--project_id",
        help="GCP project ID",
        default=default,
        callback=is_valid_project,
        required=required,
    )


def billing_project_option(default=None, required=False):
    """Generate a billing-project option, with optional default."""
    return click.option(
        "--billing-project",
        "--billing_project",
        help=(
            "GCP project ID to run the query in. "
            "This can be used to run a query using a different slot reservation "
            "than the one used by the query's default project."
        ),
        type=str,
        default=default,
        required=required,
    )


def respect_dryrun_skip_option(default=True):
    """Generate a respect_dryrun_skip option."""
    flags = {True: "--respect-dryrun-skip", False: "--ignore-dryrun-skip"}
    return click.option(
        f"{flags[True]}/{flags[False]}",
        help="Respect or ignore dry run skip configuration. "
        f"Default is {flags[default]}.",
        default=default,
    )


def no_dryrun_option(default=False):
    """Generate a skip_dryrun option."""
    return click.option(
        "--no-dryrun",
        "--no_dryrun",
        help="Skip running dryrun. " f"Default is {default}.",
        default=default,
        is_flag=True,
    )


def temp_dataset_option(
    default=f"{ConfigLoader.get('default', 'project', fallback='moz-fx-data-shared-prod')}.tmp",
):
    """Generate a temp-dataset option."""
    return click.option(
        "--temp-dataset",
        "--temp_dataset",
        "--temporary-dataset",
        "--temporary_dataset",
        default=default,
        type=TempDatasetReference.from_string,
        help="Dataset where intermediate query results will be temporarily stored, "
        "formatted as PROJECT_ID.DATASET_ID",
    )



#!/usr/bin/env python3
"""Generate udf/main_summary_scalars.sql."""

import itertools
import json
import os.path
import sys
from pathlib import Path

sys.path.append(str(Path(__file__).parent.parent.resolve()))
from bigquery_etl.util.common import snake_case  # noqa E402

SCALAR_TYPES = {"uint": "INT64", "string": "STRING", "boolean": "BOOL"}


def collect_probes(probes, schema_fields):
    """Collect scalars from probes and schema fields."""
    scalars = {"main": [], "content": [], "gpu": []}
    keyed_scalars = {"main": [], "content": [], "gpu": []}
    for probe in probes:
        history = list(itertools.chain.from_iterable(probe["history"].values()))
        record_in_processes = set().union(
            *[h["details"]["record_in_processes"] for h in history]
        )
        if probe["type"] == "scalar":
            collection = (
                keyed_scalars
                if any(h["details"]["keyed"] for h in history)
                else scalars
            )
        else:
            continue
        if record_in_processes == set(["all"]):
            record_in_processes = ["main", "content", "gpu"]
        if "all_children" in record_in_processes or "all_childs" in record_in_processes:
            record_in_processes = ["content", "gpu"]
        for p in record_in_processes:
            collection[p].append(
                (
                    snake_case(probe["name"]).replace(".", "_"),
                    SCALAR_TYPES.get(
                        history[0]["details"]["kind"], history[0]["details"]["kind"]
                    ),
                )
            )
    return scalars, keyed_scalars


def search(target, path):
    """List the field names available in target at path."""
    for key in path:
        for item in target:
            if item["name"] == key:
                target = item["fields"]
                break
    return set(x["name"] for x in target)


def collect_fields(main_schema):
    """Collect scalar fields from main ping schema."""
    scalars = {
        "main": search(main_schema, ["payload", "processes", "parent", "scalars"]),
        "content": search(main_schema, ["payload", "processes", "content", "scalars"]),
    }
    keyed_scalars = {
        "main": search(
            main_schema, ["payload", "processes", "parent", "keyed_scalars"]
        ),
        "content": search(
            main_schema, ["payload", "processes", "content", "keyed_scalars"]
        ),
    }
    return {"scalars": scalars, "keyed_scalars": keyed_scalars}


def make_field(source, target, s, schema_fields, keyed=False):
    """Build a scalar definition."""
    (name, sql_type) = s
    if name in schema_fields:
        source = "processes.%s.%s" % (source, name)
    else:
        source = "JSON_EXTRACT(additional_properties, '$.payload.processes.%s.%s')" % (
            source,
            name,
        )
        if keyed:
            kind = {"BOOL": "bool", "INT64": "string_int"}[sql_type]
            source = "udf.json_extract_%s_map(%s)" % (kind, source)
        else:
            source = "CAST(%s AS %s)" % (source, sql_type)
    return "%s AS %s_%s" % (source, target, name)


def main(root):
    """Generate udf/main_summary_scalars.sql."""
    main_schema_json = open("main.4.bq")
    main_schema = json.load(main_schema_json)
    schema_fields = collect_fields(main_schema)

    probes_json = open("all_probes.json")
    probes = json.load(probes_json).values()
    scalars, keyed_scalars = collect_probes(probes, schema_fields)
    scalar_fields = ",\n    ".join(
        [
            make_field(
                "parent.scalars", "scalar_parent", s, schema_fields["scalars"]["main"]
            )
            for s in scalars["main"]
        ]
        + [
            make_field(
                "content.scalars",
                "scalar_content",
                s,
                schema_fields["scalars"]["content"],
            )
            for s in scalars["content"]
        ]
        + [
            make_field(
                "parent.keyed_scalars",
                "scalar_parent",
                s,
                schema_fields["keyed_scalars"]["main"],
                True,
            )
            for s in keyed_scalars["main"]
        ]
        + [
            make_field(
                "content.keyed_scalars",
                "scalar_content",
                s,
                schema_fields["keyed_scalars"]["content"],
                True,
            )
            for s in keyed_scalars["content"]
        ]
    )

    scalars_sql = """
CREATE OR REPLACE FUNCTION udf.main_summary_scalars(processes ANY TYPE, additional_properties STRING) AS (
  STRUCT(
    %s
  )
);
""" % (  # noqa E501
        scalar_fields,
    )

    open(os.path.join(root, "main_summary_scalars.sql"), "w").write(scalars_sql)


if __name__ == "__main__":
    main(os.path.dirname(__file__))



"""Generate active users aggregates per app."""

import os
from enum import Enum
from pathlib import Path

import click
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.cli.utils import use_cloud_function_option
from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.util.common import render, write_sql

THIS_PATH = Path(os.path.dirname(__file__))
TABLE_NAME = os.path.basename(os.path.normpath(THIS_PATH))
BASE_NAME = "_".join(TABLE_NAME.split("_")[:-1])
DATASET_FOR_UNIONED_VIEWS = "telemetry"
CHECKS_TEMPLATE_CHANNELS = {
    "firefox_ios": [
        {
            "name": "release",
            "table": "`moz-fx-data-shared-prod.org_mozilla_ios_firefox_live.baseline_v1`",
        },
        {
            "name": "beta",
            "table": "`moz-fx-data-shared-prod.org_mozilla_ios_firefoxbeta_live.baseline_v1`",
        },
        {
            "name": "nightly",
            "table": "`moz-fx-data-shared-prod.org_mozilla_ios_fennec_live.baseline_v1`",
        },
    ],
    "focus_ios": [
        {"table": "`moz-fx-data-shared-prod.org_mozilla_ios_focus_live.baseline_v1`"}
    ],
    "focus_android": [
        {
            "name": "release",
            "table": "`moz-fx-data-shared-prod.org_mozilla_focus_live.baseline_v1`",
        },
        {
            "name": "beta",
            "table": "`moz-fx-data-shared-prod.org_mozilla_focus_beta_live.baseline_v1`",
        },
        {
            "name": "nightly",
            "table": "`moz-fx-data-shared-prod.org_mozilla_focus_nightly_live.baseline_v1`",
        },
    ],
    "klar_ios": [
        {"table": "`moz-fx-data-shared-prod.org_mozilla_ios_klar_live.baseline_v1`"}
    ],
    "klar_android" : [
        {"table" : "`moz-fx-data-shared-prod.org_mozilla_klar_live.baseline_v1`"}
    ],
}

class Browsers(Enum):
    """Enumeration with browser names and equivalent dataset names."""

    firefox_desktop = "Firefox Desktop"


@click.command()
@click.option(
    "--output-dir",
    "--output_dir",
    help="Output directory generated SQL is written to",
    type=click.Path(file_okay=False),
    default="sql",
)
@click.option(
    "--target-project",
    "--target_project",
    help="Google Cloud project ID",
    default="moz-fx-data-shared-prod",
)
@use_cloud_function_option
def generate(target_project, output_dir, use_cloud_function):
    """Generate per-app queries, views and metadata for active users and search counts aggregates.

    The parent folders will be created if not existing and existing files will be overwritten.
    """
    env = Environment(loader=FileSystemLoader(str(THIS_PATH / "templates")))
    output_dir = Path(output_dir) / target_project
    # query templates
    mobile_query_template = env.get_template("mobile_query.sql")
    desktop_query_template = env.get_template("desktop_query.sql")
    focus_android_query_template = env.get_template("focus_android_query.sql")
    # view templates
    focus_android_view_template = env.get_template("focus_android_view.sql")
    mobile_view_template = env.get_template("mobile_view.sql")
    view_template = env.get_template("view.sql")
    # metadata template
    metadata_template = "metadata.yaml"
    # schema templates
    desktop_schema_template = "desktop_schema.yaml"
    mobile_schema_template = "mobile_schema.yaml"
    # backfill template
    desktop_backfill_template = "desktop_backfill.yaml"
    # checks templates
    desktop_checks_template = env.get_template("desktop_checks.sql")
    fenix_checks_template = env.get_template("fenix_checks.sql")
    mobile_checks_template = env.get_template("mobile_checks.sql")

    for browser in Browsers:
        if browser.name == "firefox_desktop":
            query_sql = reformat(
                desktop_query_template.render(
                    app_value=browser.value,
                )
            )
            schema_template = desktop_schema_template
        elif browser.name == "focus_android":
            query_sql = reformat(
                focus_android_query_template.render(
                    project_id=target_project,
                    app_name=browser.name,
                )
            )
            schema_template = mobile_schema_template
        else:
            query_sql = reformat(
                mobile_query_template.render(
                    project_id=target_project,
                    app_value=browser.value,
                    app_name=browser.name,
                )
            )
            schema_template = mobile_schema_template

        # create checks_sql
        if browser.name == "firefox_desktop":
            checks_sql = desktop_checks_template.render(
                project_id=target_project,
                app_value=browser.value,
                app_name=browser.name,
            )
        elif browser.name == "fenix":
            checks_sql = fenix_checks_template.render(
                project_id=target_project,
                app_value=browser.value,
                app_name=browser.name,
            )
        elif browser.name in CHECKS_TEMPLATE_CHANNELS.keys():
            checks_sql = mobile_checks_template.render(
                project_id=target_project,
                app_value=browser.value,
                app_name=browser.name,
                channels=CHECKS_TEMPLATE_CHANNELS[browser.name],
            )


        # Write query SQL files.
        write_sql(
            output_dir=output_dir,
            full_table_id=f"{target_project}.{browser.name}_derived.{TABLE_NAME}",
            basename="query.sql",
            sql=query_sql,
            skip_existing=False,
        )

        # Write metadata YAML files.
        write_sql(
            output_dir=output_dir,
            full_table_id=f"{target_project}.{browser.name}_derived.{TABLE_NAME}",
            basename="metadata.yaml",
            sql=render(
                    metadata_template,
                    template_folder=THIS_PATH / "templates",
                    app_value=browser.value,
                    app_name=browser.name,
                    table_name=TABLE_NAME,
                    format=False,
                ),
            skip_existing=False,
        )

        # Write schema YAML files.
        write_sql(
            output_dir=output_dir,
            full_table_id=f"{target_project}.{browser.name}_derived.{TABLE_NAME}",
            basename="schema.yaml",
            sql=render(
                schema_template,
                template_folder=THIS_PATH / "templates",
                format=False,
            ),
            skip_existing=False,
        )

        # Write backfill YAML files.
        if browser.name == "firefox_desktop":
            write_sql(
                output_dir=output_dir,
                full_table_id=f"{target_project}.{browser.name}_derived.{TABLE_NAME}",
                basename="backfill.yaml",
                sql=render(
                    desktop_backfill_template,
                    template_folder=THIS_PATH / "templates",
                    format=False,
                ),
                skip_existing=False,
            )

        # Write checks sql files.
        write_sql(
            output_dir=output_dir,
            full_table_id=f"{target_project}.{browser.name}_derived.{TABLE_NAME}",
            basename="checks.sql",
            sql=checks_sql,
            skip_existing=False,
        )

        if browser.name == "focus_android":
            write_sql(
                output_dir=output_dir,
                full_table_id=f"{target_project}.{browser.name}.{BASE_NAME}",
                basename="view.sql",
                sql=reformat(
                    focus_android_view_template.render(
                        project_id=target_project,
                        app_name=browser.name,
                        table_name=TABLE_NAME,
                    )
                ),
                skip_existing=False,
            )
        else:
            write_sql(
                output_dir=output_dir,
                full_table_id=f"{target_project}.{browser.name}.{BASE_NAME}",
                basename="view.sql",
                sql=reformat(
                    view_template.render(
                        project_id=target_project,
                        app_name=browser.name,
                        table_name=TABLE_NAME,
                    )
                ),
                skip_existing=False,
            )



#!/usr/bin/env python3
"""clients_daily_scalar_aggregates query generator."""
import argparse
import json
import subprocess
import sys
import textwrap
import urllib.request
from pathlib import Path
from time import sleep

from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.util import probe_filters
from bigquery_etl.util.common import snake_case

sys.path.append(str(Path(__file__).parent.parent.parent.resolve()))

PROBE_INFO_SERVICE = (
    "https://probeinfo.telemetry.mozilla.org/firefox/all/main/all_probes"
)

p = argparse.ArgumentParser()
p.add_argument(
    "--agg-type",
    type=str,
    help="One of scalars/keyed_scalars/keyed_booleans",
    required=True,
)
p.add_argument(
    "--json-output",
    action="store_true",
    help="Output the result wrapped in json parseable as an XCOM",
)
p.add_argument(
    "--wait-seconds",
    type=int,
    default=0,
    help="Add a delay before executing the script to allow time for the xcom sidecar to complete startup",
)


def generate_sql(
    agg_type,
    aggregates,
    additional_queries,
    additional_partitions,
    select_clause,
    querying_table,
    json_output,
):
    """Create a SQL query for the clients_daily_scalar_aggregates dataset."""
    query = textwrap.dedent(
        f"""-- Query generated by: sql/telemetry_derived/clients_daily_scalar_aggregates.sql.py --agg-type {agg_type}
        WITH valid_build_ids AS (
            SELECT
              DISTINCT(build.build.id) AS build_id
            FROM
              `moz-fx-data-shared-prod.telemetry.buildhub2`
        ),
        filtered AS (
            SELECT
                *,
                SPLIT(application.version, '.')[OFFSET(0)] AS app_version,
                DATE(submission_timestamp) as submission_date,
                normalized_os as os,
                application.build_id AS app_build_id,
                normalized_channel AS channel
            FROM `moz-fx-data-shared-prod.telemetry_stable.main_v5`
            INNER JOIN valid_build_ids
            ON (application.build_id = build_id)
            WHERE DATE(submission_timestamp) = @submission_date
                AND normalized_channel in (
                  "release", "beta", "nightly"
                )
                AND client_id IS NOT NULL),

        {additional_queries}

        sampled_data AS (
            SELECT *
            FROM {querying_table}
            WHERE channel IN ("nightly", "beta")
                OR (channel = "release" AND os != "Windows")
                OR (channel = "release" AND
                    os = "Windows" AND
                    MOD(sample_id, @sample_size) = 0)
        ),

        -- Using `min` for when `agg_type` is `count` returns null when all rows are null
        aggregated AS (
            SELECT
                submission_date,
                sample_id,
                client_id,
                profile_group_id,
                os,
                app_version,
                app_build_id,
                channel,
                {aggregates}
            FROM sampled_data
            GROUP BY
                submission_date,
                sample_id,
                client_id,
                profile_group_id,
                os,
                app_version,
                app_build_id,
                channel
                {additional_partitions})

            {select_clause}
        """
    )

    if json_output:
        return json.dumps(query)
    else:
        return query


def _get_generic_keyed_scalar_sql(probes, value_type):
    probes_struct = []
    for probe, processes in probes.items():
        for process in processes:
            probes_struct.append(
                f"('{probe}', '{process}', payload.processes.{process}.keyed_scalars.{probe})"
            )

    probes_struct.sort()
    probes_arr = ",\n\t\t\t".join(probes_struct)

    additional_queries = f"""
        grouped_metrics AS
          (SELECT
            sample_id,
            client_id,
            profile_group_id,
            submission_date,
            os,
            app_version,
            app_build_id,
            channel,
            ARRAY<STRUCT<
                name STRING,
                process STRING,
                value ARRAY<STRUCT<key STRING, value {value_type}>>
            >>[
              {probes_arr}
            ] as metrics
          FROM filtered),

          flattened_metrics AS
            (SELECT
              sample_id,
              client_id,
              profile_group_id,
              submission_date,
              os,
              app_version,
              app_build_id,
              channel,
              metrics.name AS metric,
              metrics.process AS process,
              value.key AS key,
              value.value AS value
            FROM grouped_metrics
            CROSS JOIN UNNEST(metrics) AS metrics,
            UNNEST(metrics.value) AS value),
    """

    querying_table = "flattened_metrics"

    additional_partitions = """,
                            metric,
                            process,
                            key
    """

    return {
        "additional_queries": additional_queries,
        "additional_partitions": additional_partitions,
        "querying_table": querying_table,
    }


def get_keyed_boolean_probes_sql_string(probes):
    """Put together the subsets of SQL required to query keyed booleans."""
    sql_strings = _get_generic_keyed_scalar_sql(probes, "BOOLEAN")
    sql_strings[
        "probes_string"
    ] = """
        metric,
        key,
        process,
        SUM(CASE WHEN value = True THEN 1 ELSE 0 END) AS true_col,
        SUM(CASE WHEN value = False THEN 1 ELSE 0 END) AS false_col
    """

    sql_strings[
        "select_clause"
    ] = """
        SELECT
              sample_id,
              client_id,
              profile_group_id,
              submission_date,
              os,
              app_version,
              app_build_id,
              channel,
              ARRAY_CONCAT_AGG(ARRAY<STRUCT<
                    metric STRING,
                    metric_type STRING,
                    key STRING,
                    process STRING,
                    agg_type STRING,
                    value FLOAT64
                >>
                [
                    (metric, 'keyed-scalar-boolean', key, process, 'true', true_col),
                    (metric, 'keyed-scalar-boolean', key, process, 'false', false_col)
                ]
            ) AS scalar_aggregates
        FROM aggregated
        GROUP BY
            sample_id,
            client_id,
            profile_group_id,
            submission_date,
            os,
            app_version,
            app_build_id,
            channel
    """
    return sql_strings


def get_keyed_scalar_probes_sql_string(probes):
    """Put together the subsets of SQL required to query keyed scalars."""
    sql_strings = _get_generic_keyed_scalar_sql(probes, "INT64")
    sql_strings[
        "probes_string"
    ] = """
        metric,
        process,
        key,
        MAX(value) AS max,
        MIN(value) AS min,
        AVG(value) AS avg,
        SUM(value) AS sum,
        IF(MIN(value) IS NULL, NULL, COUNT(*)) AS count
    """

    sql_strings[
        "select_clause"
    ] = """
        SELECT
            sample_id,
            client_id,
            profile_group_id,
            submission_date,
            os,
            app_version,
            app_build_id,
            channel,
            ARRAY_CONCAT_AGG(ARRAY<STRUCT<
                metric STRING,
                metric_type STRING,
                key STRING,
                process STRING,
                agg_type STRING,
                value FLOAT64
            >>
                [
                    (metric, 'keyed-scalar', key, process, 'max', max),
                    (metric, 'keyed-scalar', key, process, 'min', min),
                    (metric, 'keyed-scalar', key, process, 'avg', avg),
                    (metric, 'keyed-scalar', key, process, 'sum', sum),
                    (metric, 'keyed-scalar', key, process, 'count', count)
                ]
        ) AS scalar_aggregates
        FROM aggregated
        GROUP BY
            sample_id,
            client_id,
            profile_group_id,
            submission_date,
            os,
            app_version,
            app_build_id,
            channel
    """
    return sql_strings


def get_scalar_probes_sql_strings(probes, scalar_type):
    """Put together the subsets of SQL required to query scalars or booleans."""
    if scalar_type == "keyed_scalars":
        return get_keyed_scalar_probes_sql_string(probes["keyed"])

    if scalar_type == "keyed_booleans":
        return get_keyed_boolean_probes_sql_string(probes["keyed_boolean"])

    probe_structs = []
    for probe, processes in probes["scalars"].items():
        for process in processes:
            probe_structs.append(
                (
                    f"('{probe}', 'scalar', '', '{process}', 'max', "
                    f"max(CAST(payload.processes.{process}.scalars.{probe} AS INT64)))"
                )
            )
            probe_structs.append(
                (
                    f"('{probe}', 'scalar', '', '{process}', 'avg', "
                    f"avg(CAST(payload.processes.{process}.scalars.{probe} AS INT64)))"
                )
            )
            probe_structs.append(
                (
                    f"('{probe}', 'scalar', '', '{process}', 'min', "
                    f"min(CAST(payload.processes.{process}.scalars.{probe} AS INT64)))"
                )
            )
            probe_structs.append(
                (
                    f"('{probe}', 'scalar', '', '{process}', 'sum', "
                    f"sum(CAST(payload.processes.{process}.scalars.{probe} AS INT64)))"
                )
            )
            probe_structs.append(
                f"('{probe}', 'scalar', '', '{process}', 'count', IF(MIN(payload.processes.{process}.scalars.{probe}) IS NULL, NULL, COUNT(*)))"
            )

    for probe, processes in probes["booleans"].items():
        for process in processes:
            probe_structs.append(
                (
                    f"('{probe}', 'boolean', '', '{process}', 'false', "
                    f"SUM(case when payload.processes.{process}.scalars.{probe} = False "
                    "THEN 1 ELSE 0 END))"
                )
            )
            probe_structs.append(
                (
                    f"('{probe}', 'boolean', '', '{process}', 'true', "
                    f"SUM(case when payload.processes.{process}.scalars.{probe} = True "
                    "THEN 1 ELSE 0 END))"
                )
            )

    probe_structs.sort()
    probes_arr = ",\n\t\t\t".join(probe_structs)
    probes_string = f"""
            ARRAY<STRUCT<
                metric STRING,
                metric_type STRING,
                key STRING,
                process STRING,
                agg_type STRING,
                value FLOAT64
            >> [
                {probes_arr}
            ] AS scalar_aggregates
    """

    select_clause = """
        SELECT *
        FROM aggregated
    """

    return {"probes_string": probes_string, "select_clause": select_clause}


def save_scalars_by_type(scalars_dict, scalar, process):
    """Add scalars to scalars_dict based on their type."""
    if scalars_dict is None:
        return

    processes = scalars_dict.setdefault(scalar, set())
    processes.add(process)
    scalars_dict[scalar] = processes


def filter_scalars_dict(scalars_dict, required_probes):
    """Filter out scalars that are not required."""
    return {
        scalar: process
        for scalar, process in scalars_dict.items()
        if scalar in required_probes
    }


def get_scalar_probes(scalar_type):
    """Find all scalar probes in main summary.

    Note: that non-integer scalar probes are not included.
    """
    project = "moz-fx-data-shared-prod"
    main_summary_scalars = {}
    main_summary_record_scalars = {}
    main_summary_boolean_record_scalars = {}
    main_summary_boolean_scalars = {}

    process = subprocess.Popen(
        [
            "bq",
            "show",
            "--schema",
            "--format=json",
            f"{project}:telemetry_stable.main_v5",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = process.communicate()
    if process.returncode > 0:
        raise Exception(
            f"Call to bq exited non-zero: {process.returncode}", stdout, stderr
        )
    main_summary_schema = json.loads(stdout)

    scalars_fields = []
    for field in main_summary_schema:
        if field["name"] != "payload":
            continue

        for payload_field in field["fields"]:
            if payload_field["name"] == "processes":
                for processes_field in payload_field["fields"]:
                    if processes_field["name"] in ["parent", "content", "gpu"]:
                        process_field = processes_field["name"]
                        for type_field in processes_field["fields"]:
                            if type_field["name"] == scalar_type:
                                scalars_fields.append(
                                    {"scalars": type_field, "process": process_field}
                                )
                                break

    if len(scalars_fields) == 0:
        return

    for scalars_and_process in scalars_fields:
        for scalar in scalars_and_process["scalars"].get("fields", {}):
            scalars_dict = None
            if "name" not in scalar:
                continue

            if scalar.get("type", "") == "INTEGER":
                scalars_dict = main_summary_scalars
            elif scalar.get("type", "") == "BOOLEAN":
                scalars_dict = main_summary_boolean_scalars
            elif scalar.get("type", "") == "RECORD":
                if scalar["fields"][1]["type"] == "BOOLEAN":
                    scalars_dict = main_summary_boolean_record_scalars
                else:
                    scalars_dict = main_summary_record_scalars

            save_scalars_by_type(
                scalars_dict, scalar["name"], scalars_and_process["process"]
            )

    # Find the intersection between relevant scalar probes
    # and those that exist in main summary
    with urllib.request.urlopen(PROBE_INFO_SERVICE) as url:
        data = json.loads(url.read())
        excluded_probes = probe_filters.get_etl_excluded_probes_quickfix("desktop")
        scalar_probes = (
            set(
                [
                    snake_case(x.replace("scalar/", ""))
                    for x in data.keys()
                    if x.startswith("scalar/")
                ]
            )
            - excluded_probes
        )

        return {
            "scalars": filter_scalars_dict(main_summary_scalars, scalar_probes),
            "booleans": filter_scalars_dict(
                main_summary_boolean_scalars, scalar_probes
            ),
            "keyed": filter_scalars_dict(main_summary_record_scalars, scalar_probes),
            "keyed_boolean": filter_scalars_dict(
                main_summary_boolean_record_scalars, scalar_probes
            ),
        }


def main(argv, out=print):
    """Print a clients_daily_scalar_aggregates query to stdout."""
    opts = vars(p.parse_args(argv[1:]))
    sql_string = ""

    if opts["agg_type"] in ("scalars", "keyed_scalars", "keyed_booleans"):
        scalar_type = (
            opts["agg_type"] if (opts["agg_type"] == "scalars") else "keyed_scalars"
        )
        scalar_probes = get_scalar_probes(scalar_type)
        sql_string = get_scalar_probes_sql_strings(scalar_probes, opts["agg_type"])
    else:
        raise ValueError(
            "agg-type must be one of scalars, keyed_scalars, keyed_booleans"
        )

    sleep(opts["wait_seconds"])
    out(
        reformat(
            generate_sql(
                opts["agg_type"],
                sql_string["probes_string"],
                sql_string.get("additional_queries", ""),
                sql_string.get("additional_partitions", ""),
                sql_string["select_clause"],
                sql_string.get("querying_table", "filtered"),
                opts["json_output"],
            )
        )
    )


if __name__ == "__main__":
    main(sys.argv)



from pathlib import Path

from bigquery_etl.config import ConfigLoader

TEST_DIR = Path(__file__).parent


class TestConfig:
    def test_config_loader_set_project(self):
        ConfigLoader.set_project_dir(TEST_DIR / "data")
        assert ConfigLoader.project_dir == TEST_DIR / "data"

    def test_config_loader_get(self):
        ConfigLoader.set_project_dir(TEST_DIR / "data")

        assert "function" in ConfigLoader.get("dry_run")
        assert "skip" in ConfigLoader.get("dry_run")

        assert (
            ConfigLoader.get("default", "test_project")
            == "bigquery-etl-integration-test"
        )

        assert len(ConfigLoader.get("dry_run", "skip")) > 0

        assert "dry_run" in ConfigLoader.get()

    def test_config_loader_get_non_existing(self):
        ConfigLoader.set_project_dir(TEST_DIR / "data")

        assert ConfigLoader.get("non_existing") is None
        assert ConfigLoader.get("dry_run", "non_existing") is None
        assert ConfigLoader.get("non_existing", fallback=[]) == []
        assert ConfigLoader.get("dry_run", "foo", fallback=123) == 123



"""Generate Materialized Views and aggregate queries for event monitoring."""

import os
import re

from collections import namedtuple, OrderedDict
from datetime import datetime
from pathlib import Path
from typing import List, Set

import requests

from bigquery_etl.config import ConfigLoader
from bigquery_etl.schema.stable_table_schema import get_stable_table_schemas
from sql_generators.glean_usage.common import (
    GleanTable,
    get_app_info,
    get_table_dir,
    render,
    table_names_from_baseline,
    write_sql,
)

TARGET_TABLE_ID = "event_monitoring_live_v1"
TARGET_DATASET_CROSS_APP = "monitoring"
PREFIX = "event_monitoring"
PATH = Path(os.path.dirname(__file__))
METRICS_INFO_URL = "https://probeinfo.telemetry.mozilla.org/glean/{app_name}/metrics"


class EventMonitoringLive(GleanTable):
    """Represents the generated materialized view for event monitoring."""

    def __init__(self) -> None:
        """Initialize materialized view generation."""
        self.per_app_id_enabled = True
        self.per_app_enabled = False
        self.across_apps_enabled = True
        self.prefix = PREFIX
        self.target_table_id = TARGET_TABLE_ID
        self.custom_render_kwargs = {}
        self.base_table_name = "events_v1"

    def _get_prod_datasets_with_event(self) -> List[str]:
        """Get glean datasets with an events table in generated schemas."""
        return [
            s.bq_dataset_family
            for s in get_stable_table_schemas()
            if s.schema_id == "moz://mozilla.org/schemas/glean/ping/1"
            and s.bq_table == "events_v1"
        ]

    def _get_tables_with_events(self, v1_name: str, bq_dataset_name: str) -> Set[str]:
        """Get tables for the given app that receive event type metrics."""
        pings = set()
        resp = requests.get(METRICS_INFO_URL.format(app_name=v1_name))
        resp.raise_for_status()
        metrics_json = resp.json()

        for _, metric in metrics_json.items():
            if metric.get("type", None) == "event":
                latest_history = metric.get("history", [])[-1]
                pings.update(latest_history.get("send_in_pings", []))

        if bq_dataset_name in self._get_prod_datasets_with_event():
            pings.add("events")

        return pings

    def generate_per_app_id(
        self,
        project_id,
        baseline_table,
        output_dir=None,
        use_cloud_function=True,
        app_info=[],
        parallelism=8,
        id_token=None,
    ):
        # Get the app ID from the baseline_table name.
        # This is what `common.py` also does.
        app_id = re.sub(r"_stable\..+", "", baseline_table)
        app_id = ".".join(app_id.split(".")[1:])

        # Skip any not-allowed app.
        if app_id in ConfigLoader.get(
            "generate", "glean_usage", "events_monitoring", "skip_apps", fallback=[]
        ):
            return

        tables = table_names_from_baseline(baseline_table, include_project_id=False)

        init_filename = f"{self.target_table_id}.materialized_view.sql"
        metadata_filename = f"{self.target_table_id}.metadata.yaml"

        table = tables[f"{self.prefix}"]
        dataset = tables[self.prefix].split(".")[-2].replace("_derived", "")

        events_table_overwrites = ConfigLoader.get(
            "generate", "glean_usage", "events_monitoring", "events_tables", fallback={}
        )

        app_name = [
            app_dataset["app_name"]
            for _, app in get_app_info().items()
            for app_dataset in app
            if dataset == app_dataset["bq_dataset_family"]
        ][0]

        if app_name in events_table_overwrites:
            events_tables = events_table_overwrites[app_name]
        else:
            v1_name = [
                app_dataset["v1_name"]
                for _, app in get_app_info().items()
                for app_dataset in app
                if dataset == app_dataset["bq_dataset_family"]
            ][0]
            events_tables = self._get_tables_with_events(v1_name, dataset)
            events_tables = [
                f"{ping.replace('-', '_')}_v1"
                for ping in events_tables
                if ping
                not in ConfigLoader.get(
                    "generate", "glean_usage", "events_monitoring", "skip_pings"
                )
            ]

        if len(events_tables) == 0:
            return

        render_kwargs = dict(
            header="-- Generated via bigquery_etl.glean_usage\n",
            header_yaml="---\n# Generated via bigquery_etl.glean_usage\n",
            project_id=project_id,
            derived_dataset=tables[self.prefix].split(".")[-2],
            dataset=dataset,
            current_date=datetime.today().strftime("%Y-%m-%d"),
            app_name=[
                app_dataset["canonical_app_name"]
                for _, app in get_app_info().items()
                for app_dataset in app
                if dataset == app_dataset["bq_dataset_family"]
            ][0],
            events_tables=sorted(events_tables),
        )

        render_kwargs.update(self.custom_render_kwargs)
        render_kwargs.update(tables)

        # generated files to update
        Artifact = namedtuple("Artifact", "table_id basename sql")
        artifacts = []

        init_sql = render(
            init_filename, template_folder=PATH / "templates", **render_kwargs
        )
        metadata = render(
            metadata_filename,
            template_folder=PATH / "templates",
            format=False,
            **render_kwargs,
        )
        artifacts.append(Artifact(table, "metadata.yaml", metadata))

        skip_existing_artifact = self.skip_existing(output_dir, project_id)

        if output_dir:
            artifacts.append(Artifact(table, "materialized_view.sql", init_sql))

            for artifact in artifacts:
                destination = (
                    get_table_dir(output_dir, artifact.table_id) / artifact.basename
                )
                skip_existing = str(destination) in skip_existing_artifact

                write_sql(
                    output_dir,
                    artifact.table_id,
                    artifact.basename,
                    artifact.sql,
                    skip_existing=skip_existing,
                )

    def generate_across_apps(
        self, project_id, apps, output_dir=None, use_cloud_function=True, parallelism=8
    ):
        """Generate a query across all apps."""
        if not self.across_apps_enabled:
            return

        aggregate_table = "event_monitoring_aggregates_v1"
        target_view_name = "_".join(self.target_table_id.split("_")[:-1])

        events_table_overwrites = ConfigLoader.get(
            "generate", "glean_usage", "events_monitoring", "events_tables", fallback={}
        )

        event_tables_per_dataset = OrderedDict()

        # Skip any not-allowed app.
        skip_apps = ConfigLoader.get(
            "generate", "glean_usage", "events_monitoring", "skip_apps", fallback=[]
        )

        for app in apps:
            for app_dataset in app:
                if app_dataset["app_name"] in skip_apps:
                    continue

                dataset = app_dataset["bq_dataset_family"]
                app_name = [
                    app_dataset["app_name"]
                    for _, app in get_app_info().items()
                    for app_dataset in app
                    if dataset == app_dataset["bq_dataset_family"]
                ][0]

                if app_name in events_table_overwrites:
                    event_tables_per_dataset[dataset] = events_table_overwrites[
                        app_name
                    ]
                else:
                    v1_name = [
                        app_dataset["v1_name"]
                        for _, app in get_app_info().items()
                        for app_dataset in app
                        if dataset == app_dataset["bq_dataset_family"]
                    ][0]
                    event_tables = [
                        f"{ping.replace('-', '_')}_v1"
                        for ping in self._get_tables_with_events(
                            v1_name, app_dataset["bq_dataset_family"]
                        )
                        if ping
                        not in ConfigLoader.get(
                            "generate", "glean_usage", "events_monitoring", "skip_pings"
                        )
                    ]

                    if len(event_tables) > 0:
                        event_tables_per_dataset[dataset] = sorted(event_tables)

        render_kwargs = dict(
            header="-- Generated via bigquery_etl.glean_usage\n",
            header_yaml="---\n# Generated via bigquery_etl.glean_usage\n",
            project_id=project_id,
            target_view=f"{TARGET_DATASET_CROSS_APP}.{target_view_name}",
            table=target_view_name,
            target_table=f"{TARGET_DATASET_CROSS_APP}_derived.{aggregate_table}",
            apps=apps,
            prod_datasets=self._get_prod_datasets_with_event(),
            event_tables_per_dataset=event_tables_per_dataset,
        )
        render_kwargs.update(self.custom_render_kwargs)

        skip_existing_artifacts = self.skip_existing(output_dir, project_id)

        Artifact = namedtuple("Artifact", "table_id basename sql")

        query_filename = f"{aggregate_table}.query.sql"
        query_sql = render(
            query_filename, template_folder=PATH / "templates", **render_kwargs
        )
        metadata = render(
            f"{aggregate_table}.metadata.yaml",
            template_folder=PATH / "templates",
            format=False,
            **render_kwargs,
        )
        table = f"{project_id}.{TARGET_DATASET_CROSS_APP}_derived.{aggregate_table}"

        view_sql = render(
            "event_monitoring_live.view.sql",
            template_folder=PATH / "templates",
            **render_kwargs,
        )
        view_metadata = render(
            "event_monitoring_live.metadata.yaml",
            template_folder=PATH / "templates",
            format=False,
            **render_kwargs,
        )

        view = f"{project_id}.{TARGET_DATASET_CROSS_APP}.{target_view_name}"
        if output_dir:
            artifacts = [
                Artifact(table, "metadata.yaml", metadata),
                Artifact(table, "query.sql", query_sql),
                Artifact(view, "metadata.yaml", view_metadata),
                Artifact(view, "view.sql", view_sql),
            ]

            for artifact in artifacts:
                destination = (
                    get_table_dir(output_dir, artifact.table_id) / artifact.basename
                )
                skip_existing = destination in skip_existing_artifacts

                write_sql(
                    output_dir,
                    artifact.table_id,
                    artifact.basename,
                    artifact.sql,
                    skip_existing=skip_existing,
                )



"""bigquery-etl CLI UDF command."""

import copy
import os
import re
import shutil
import string
import sys
from fnmatch import fnmatchcase
from glob import glob
from pathlib import Path

import pytest
import rich_click as click
import yaml

from ..cli.format import format
from ..cli.utils import (
    is_authenticated,
    is_valid_project,
    project_id_option,
    sql_dir_option,
)
from ..config import ConfigLoader
from ..docs import validate as validate_docs
from ..format_sql.formatter import reformat
from ..routine import publish_routines
from ..routine.parse_routine import PROCEDURE_FILE, UDF_FILE
from ..util.common import project_dirs

ROUTINE_NAME_RE = re.compile(r"^(?P<dataset>[a-zA-z0-9_]+)\.(?P<name>[a-zA-z0-9_]+)$")
ROUTINE_DATASET_RE = re.compile(r"^(?P<dataset>[a-zA-z0-9_]+)$")
ROUTINE_FILE_RE = re.compile(
    r"^.*/([a-zA-Z0-9-]+)/([a-zA-Z0-9_]+)/([a-zA-Z0-9_]+)/"
    r"(udf\.sql|stored_procedure\.sql)$"
)


def _routines_matching_name_pattern(pattern, sql_path, project_id):
    """Return paths to routines matching the name pattern."""
    sql_path = Path(sql_path)
    if project_id is not None:
        sql_path = sql_path / project_id

    all_sql_files = map(Path, glob(f"{sql_path}/**/*.sql", recursive=True))
    routine_files = []

    for sql_file in all_sql_files:
        match = ROUTINE_FILE_RE.match(str(sql_file))
        if match:
            project = match.group(1)
            dataset = match.group(2)
            routine_name = match.group(3)
            routine_name = f"{project}.{dataset}.{routine_name}"
            if fnmatchcase(routine_name, f"*{pattern}"):
                routine_files.append(sql_file)
            elif project_id and fnmatchcase(routine_name, f"{project_id}.{pattern}"):
                routine_files.append(sql_file)

    return routine_files


def get_project_id(ctx, project_id=None):
    """Return the project id with the option flag taking priority."""
    if project_id:
        return project_id
    default_project = ctx.obj["DEFAULT_PROJECT"]
    if default_project and is_valid_project(ctx, None, default_project):
        return default_project
    click.echo(
        "Please specify a project_id e.g. --project_id=moz-fx-data-shared-prod",
        err=True,
    )
    sys.exit(1)


@click.group(help="Commands for managing routines for internal use.")
@click.pass_context
def routine(ctx):
    """Create the CLI group for the routine command."""
    ctx.ensure_object(dict)
    ctx.obj["DEFAULT_PROJECT"] = ConfigLoader.get("default", "project")


@click.group(help="Commands for managing public mozfun routines.")
@click.pass_context
def mozfun(ctx):
    """Create the CLI group for the mozfun command."""
    ctx.ensure_object(dict)
    ctx.obj["DEFAULT_PROJECT"] = ConfigLoader.get("routine", "project")


@routine.command(
    help="""Create a new routine. Specify whether the routine is a UDF or
    stored procedure by adding a --udf or --stored_prodecure flag.

    Examples:

    \b
    # Create a UDF
    ./bqetl routine create --udf udf.array_slice

    \b
    # Create a stored procedure
    ./bqetl routine create --stored_procedure udf.events_daily

    \b
    # Create a UDF in a project other than shared-prod
    ./bqetl routine create --udf udf.active_last_week --project=moz-fx-data-marketing-prod
    """
)
@click.argument("name")
@sql_dir_option
@project_id_option()
@click.option("--udf", "-u", is_flag=True, help="Create a new UDF", default=False)
@click.option(
    "--stored_procedure",
    "--stored-procedure",
    "-p",
    is_flag=True,
    help="Create a new stored procedure",
    default=False,
)
@click.pass_context
def create(ctx, name, sql_dir, project_id, udf, stored_procedure):
    """CLI command for creating a new routine."""
    if (udf is False and stored_procedure is False) or (udf and stored_procedure):
        click.echo(
            "Please specify if new routine is a UDF or stored procedure by adding "
            "either a --udf or --stored_prodecure flag: "
            "bqetl routine create <dataset>.<name> --udf/--stored_procedure",
            err=True,
        )
        sys.exit(1)

    project_id = get_project_id(ctx, project_id)

    # create directory structure
    try:
        match = ROUTINE_NAME_RE.match(name)
        name = match.group("name")
        dataset = match.group("dataset")
    except AttributeError:
        click.echo("New routine must be named like: <dataset>.<routine_name>")
        sys.exit(1)

    routine_path = Path(sql_dir) / project_id / dataset / name
    routine_path.mkdir(parents=True)

    assert_udf_qualifier = "" if project_id == "mozfun" else "mozfun."

    # create SQL file with UDF definition
    if udf:
        routine_file = routine_path / UDF_FILE
        routine_file.write_text(
            reformat(
                f"""
                -- Definition for {dataset}.{name}
                -- For more information on writing UDFs see:
                -- https://docs.telemetry.mozilla.org/cookbooks/bigquery/querying.html
                CREATE OR REPLACE FUNCTION {dataset}.{name}()
                RETURNS BOOLEAN AS (
                    TRUE
                );

                -- Tests
                SELECT {assert_udf_qualifier}assert.true({dataset}.{name}())
                """
            )
            + "\n"
        )
    elif stored_procedure:
        stored_procedure_file = routine_path / PROCEDURE_FILE
        stored_procedure_file.write_text(
            reformat(
                f"""
                -- Definition for {dataset}.{name}
                CREATE OR REPLACE PROCEDURE {dataset}.{name}()
                BEGIN

                END;

                -- Tests
                SELECT {assert_udf_qualifier}assert.true({dataset}.{name}())
                """
            )
            + "\n"
        )

    # create default metadata.yaml
    metadata_file = routine_path / "metadata.yaml"
    metadata = {
        "friendly_name": string.capwords(name.replace("_", " ")),
        "description": "Please provide a description for the routine",
    }
    metadata_file.write_text(yaml.dump(metadata))

    # create stub README.md
    readme_file = routine_path / "README.md"
    readme_file.write_text(
        (
            "\n".join(
                map(
                    lambda l: l.lstrip(),
                    """
            <!--
            This is a short README for your routine, you can add any extra
            documentation or examples that a user might want to see when
            viewing the documentation at https://mozilla.github.io/bigquery-etl

            You can embed an SQL file into your README using the following
            syntax:

            @sql(../examples/fenix_app_info.sql)
            -->
            """.split(
                        "\n"
                    ),
                )
            )
        )
        + "\n"
    )


mozfun.add_command(copy.copy(create))
mozfun.commands[
    "create"
].help = """
Create a new mozfun routine. Specify whether the routine is a UDF or
stored procedure by adding a --udf or --stored_prodecure flag. UDFs
are added to the `mozfun` project.

Examples:

\b
# Create a UDF
./bqetl mozfun create --udf bytes.zero_right

\b
# Create a stored procedure
./bqetl mozfun create --stored_procedure event_analysis.events_daily
"""


@routine.command(
    help="""Get routine information.

    Examples:

    \b
    # Get information about all internal routines in a specific dataset
    ./bqetl routine info udf.*

    \b
    # Get usage information of specific routine
    ./bqetl routine info --usages udf.get_key
    """
)
@click.argument("name", required=False)
@sql_dir_option
@project_id_option()
@click.option("--usages", "-u", is_flag=True, help="Show routine usages", default=False)
@click.pass_context
def info(ctx, name, sql_dir, project_id, usages):
    """CLI command for returning information about routines."""
    project_id = get_project_id(ctx, project_id)

    if name is None:
        name = "*.*"

    routine_files = _routines_matching_name_pattern(name, sql_dir, project_id)

    for routine_file in routine_files:
        routine_file_path = Path(routine_file)
        routine_name = routine_file_path.parent.name
        routine_dataset = routine_file_path.parent.parent.name
        routine_project = routine_file_path.parent.parent.parent.name

        try:
            metadata = yaml.safe_load(open(routine_file_path.parent / "metadata.yaml"))
        except FileNotFoundError:
            metadata = None

        click.secho(f"{routine_project}.{routine_dataset}.{routine_name}", bold=True)
        click.echo(f"path: {routine_file_path}")

        if metadata is None:
            click.echo("No metadata")
        else:
            click.echo(f"description: {metadata['description']}")

        no_usages = True
        if usages:
            # find routine usages in SQL files
            click.echo("usages: ")
            sql_files = [
                p
                for project in project_dirs()
                for p in map(Path, glob(f"{project}/**/*.sql", recursive=True))
            ]
            for sql_file in sql_files:
                if f"{routine_dataset}.{routine_name}" in sql_file.read_text():
                    no_usages = False
                    click.echo(f"  {sql_file}")

            if no_usages:
                click.echo("  No usages.")

        click.echo("")


mozfun.add_command(copy.copy(info))
mozfun.commands[
    "info"
].help = """Get mozfun routine information.

Examples:

\b
# Get information about all internal routines in a specific dataset
./bqetl mozfun info hist.*

\b
# Get usage information of specific routine
./bqetl mozfun info --usages hist.mean
"""


@routine.command(
    help="""Validate formatting of routines and run tests.

    Examples:

    \b
    # Validate all routines
    ./bqetl routine validate

    \b
    # Validate selected routines
    ./bqetl routine validate udf.*
    """,
)
@click.argument("name", required=False)
@sql_dir_option
@project_id_option()
@click.option(
    "--docs-only",
    "--docs_only",
    default=False,
    is_flag=True,
    help="Only validate docs.",
)
@click.pass_context
def validate(ctx, name, sql_dir, project_id, docs_only):
    """Validate routines by formatting and running tests."""
    project_id = get_project_id(ctx, project_id)

    if name is None:
        name = "*.*"

    routine_files = _routines_matching_name_pattern(name, sql_dir, project_id)

    ctx.invoke(validate_docs, project_dirs=project_dirs(project_id))

    if not docs_only:
        for routine_file in routine_files:
            ctx.invoke(format, paths=[str(routine_file.parent)], check=True)
            pytest.main([str(routine_file.parent)])


mozfun.add_command(copy.copy(validate))
mozfun.commands[
    "validate"
].help = """Validate formatting of mozfun routines and run tests.

Examples:

\b
# Validate all routines
./bqetl mozfun validate

\b
# Validate selected routines
./bqetl mozfun validate hist.*
"""


@routine.command(
    help="""Publish routines to BigQuery. Requires service account access.

    Examples:

    \b
    # Publish all routines
    ./bqetl routine publish

    \b
    # Publish selected routines
    ./bqetl routine validate udf.*
    """,
)
@click.argument("name", required=False)
@project_id_option()
@click.option(
    "--dependency-dir",
    "--dependency_dir",
    default=ConfigLoader.get("routine", "dependency_dir"),
    help="The directory JavaScript dependency files for UDFs are stored.",
)
@click.option(
    "--gcs-bucket",
    "--gcs_bucket",
    default=ConfigLoader.get("routine", "publish", "gcs_bucket"),
    help="The GCS bucket where dependency files are uploaded to.",
)
@click.option(
    "--gcs-path",
    "--gcs_path",
    default=ConfigLoader.get("routine", "publish", "gcs_path"),
    help="The GCS path in the bucket where dependency files are uploaded to.",
)
@click.option(
    "--dry_run/--no_dry_run", "--dry-run/--no-dry-run", help="Dry run publishing udfs."
)
@click.pass_context
def publish(ctx, name, project_id, dependency_dir, gcs_bucket, gcs_path, dry_run):
    """Publish routines."""
    project_id = get_project_id(ctx, project_id)

    public = False

    if not is_authenticated():
        click.echo("User needs to be authenticated to publish routines.", err=True)
        sys.exit(1)

    click.echo(f"Publish routines to {project_id}")
    # NOTE: this will only publish to a single project
    for target in project_dirs(project_id):
        publish_routines.publish(
            target,
            project_id,
            dependency_dir,
            gcs_bucket,
            gcs_path,
            public,
            pattern=name,
            dry_run=dry_run,
        )
        click.echo(f"Published routines to {project_id}")


mozfun.add_command(copy.copy(publish))
mozfun.commands[
    "publish"
].help = """Publish mozfun routines. This command is used
by Airflow only."""


@routine.command(
    help="""Rename routine or routine dataset. Replaces all usages in queries with
    the new name.

    Examples:

    \b
    # Rename routine
    ./bqetl routine rename udf.array_slice udf.list_slice

    \b
    # Rename routine matching a specific pattern
    ./bqetl routine rename udf.array_* udf.list_*
    """,
)
@click.argument("name", required=True)
@click.argument("new_name", required=True)
@sql_dir_option
@project_id_option()
@click.pass_context
def rename(ctx, name, new_name, sql_dir, project_id):
    """Rename routines based on pattern."""
    project_id = get_project_id(ctx, project_id)

    routine_files = _routines_matching_name_pattern(name, sql_dir, project_id)

    if ROUTINE_NAME_RE.match(new_name) and len(routine_files) <= 1:
        # rename routines
        match = ROUTINE_NAME_RE.match(new_name)
        new_routine_name = match.group("name")
        new_routine_dataset = match.group("dataset")
    elif ROUTINE_DATASET_RE.match(new_name):
        # rename routines dataset
        match = ROUTINE_DATASET_RE.match(new_name)
        new_routine_name = None
        new_routine_dataset = match.group("dataset")
    else:
        click.echo("Invalid rename naming patterns.")
        sys.exit(1)

    for routine_file in routine_files:
        # move to new directories
        old_full_routine_name = (
            f"{routine_file.parent.parent.name}.{routine_file.parent.name}"
        )

        if new_routine_name and new_routine_dataset:
            source = routine_file.parent
            destination = (
                routine_file.parent.parent.parent
                / new_routine_dataset
                / new_routine_name
            )
            new_full_routine_name = f"{new_routine_dataset}.{new_routine_name}"
        else:
            source = routine_file.parent.parent
            destination = routine_file.parent.parent.parent / new_routine_dataset
            new_full_routine_name = f"{new_routine_dataset}.{routine_file.parent.name}"

        if source.exists():
            os.makedirs(destination.parent, exist_ok=True)
            shutil.move(source, destination)

        # replace usages
        all_sql_files = [
            p
            for project in project_dirs()
            for p in map(Path, glob(f"{project}/**/*.sql", recursive=True))
        ]

        for sql_file in all_sql_files:
            sql = sql_file.read_text()

            replaced_sql = sql.replace(
                f"{old_full_routine_name}(", f"{new_full_routine_name}("
            )
            sql_file.write_text(replaced_sql)

        click.echo(f"Renamed {old_full_routine_name} to {new_full_routine_name}")


mozfun.add_command(copy.copy(rename))
mozfun.commands[
    "rename"
].help = """Rename mozfun routine or mozfun routine dataset.
Replaces all usages in queries with the new name.

Examples:

\b
# Rename routine
./bqetl mozfun rename hist.extract hist.ext

\b
# Rename routine matching a specific pattern
./bqetl mozfun rename *.array_* *.list_*

\b
# Rename routine dataset
./bqetl mozfun rename hist.* histogram.*
"""



"""Parsing of metadata yaml files."""

import enum
import os
import re
import string
from datetime import date
from pathlib import Path
from typing import Any, Dict, List, Optional

import attr
import cattrs
import yaml
from google.cloud import bigquery

from bigquery_etl.query_scheduling.utils import is_email, is_email_or_github_identity

METADATA_FILE = "metadata.yaml"
DATASET_METADATA_FILE = "dataset_metadata.yaml"
DEFAULT_WORKGROUP_ACCESS = [
    dict(role="roles/bigquery.dataViewer", members=["workgroup:mozilla-confidential"])
]
DEFAULT_TABLE_WORKGROUP_ACCESS = DEFAULT_WORKGROUP_ACCESS


class Literal(str):
    """Represents a YAML literal."""

    pass


def literal_presenter(dumper, data):
    """Literal representer for YAML output."""
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")


yaml.add_representer(Literal, literal_presenter)


class PartitionType(enum.Enum):
    """Represents BigQuery table partition types."""

    HOUR = "hour"
    DAY = "day"
    MONTH = "month"
    YEAR = "year"

    @property
    def bigquery_type(self):
        """Map to the BigQuery data type."""
        d = {
            "hour": bigquery.TimePartitioningType.HOUR,
            "day": bigquery.TimePartitioningType.DAY,
            "month": bigquery.TimePartitioningType.MONTH,
            "year": bigquery.TimePartitioningType.YEAR,
        }
        return d[self.value]


@attr.s(auto_attribs=True)
class PartitionMetadata:
    """Metadata for defining BigQuery table partitions."""

    type: PartitionType
    field: Optional[str] = attr.ib(None)
    require_partition_filter: bool = attr.ib(True)
    expiration_days: Optional[float] = attr.ib(None)

    @property
    def expiration_ms(self) -> Optional[float]:
        """Convert partition expiration from days to milliseconds."""
        if self.expiration_days is None:
            return None

        return int(self.expiration_days * 86400000)


@attr.s(auto_attribs=True)
class PartitionRange:
    """Metadata for defining the partition range."""

    start: int
    end: int
    interval: int


@attr.s(auto_attribs=True)
class RangePartitionMetadata:
    """Metadata for defining range partitioned tables."""

    range: PartitionRange
    field: Optional[str] = attr.ib(None)


@attr.s(auto_attribs=True)
class ClusteringMetadata:
    """Metadata for defining BigQuery table clustering."""

    fields: List[str]


@attr.s(auto_attribs=True)
class BigQueryMetadata:
    """
    Metadata related to BigQuery configurations for the query.

    For example, partitioning or clustering of the destination table.
    """

    time_partitioning: Optional[PartitionMetadata] = attr.ib(None)
    range_partitioning: Optional[RangePartitionMetadata] = attr.ib(None)
    clustering: Optional[ClusteringMetadata] = attr.ib(None)


@attr.s(auto_attribs=True)
class SchemaDerivedMetadata:
    """Metadata specifying parent schema."""

    table: List[str]
    # list of excluded columns
    exclude: Optional[List[str]] = attr.ib(None)


@attr.s(auto_attribs=True)
class SchemaMetadata:
    """Metadata related to additional schema information."""

    derived_from: List[SchemaDerivedMetadata]


@attr.s(auto_attribs=True)
class WorkgroupAccessMetadata:
    """Workgroup access metadata."""

    role: str
    members: List[str]


class ExternalDataFormat(enum.Enum):
    """Represents the external types fo data that are supported to be integrated."""

    GOOGLE_SHEETS = "google_sheets"
    CSV = "csv"


@attr.s(auto_attribs=True)
class ExternalDataMetadata:
    """Metadata for specifying external data."""

    format: ExternalDataFormat
    source_uris: List[str]
    options: Optional[Dict[str, Any]] = attr.ib(None)


@attr.s(auto_attribs=True)
class MonitoringMetadata:
    """Metadata for specifying observability and monitoring configuration."""

    enabled: bool = attr.ib(True)
    collection: Optional[str] = attr.ib(None)
    partition_column: Optional[str] = attr.ib(None)
    partition_column_set: bool = attr.ib(False)


@attr.s(auto_attribs=True)
class Metadata:
    """
    Representation of a table or view Metadata configuration.

    Uses attrs to simplify the class definition and provide validation.
    Docs: https://www.attrs.org
    """

    friendly_name: str = attr.ib()
    description: str = attr.ib()
    owners: List[str] = attr.ib()
    labels: Dict = attr.ib({})
    scheduling: Optional[Dict] = attr.ib({})
    bigquery: Optional[BigQueryMetadata] = attr.ib(None)
    schema: Optional[SchemaMetadata] = attr.ib(None)
    workgroup_access: Optional[List[WorkgroupAccessMetadata]] = attr.ib(None)
    references: Dict = attr.ib({})
    external_data: Optional[ExternalDataMetadata] = attr.ib(None)
    deprecated: bool = attr.ib(False)
    deletion_date: Optional[date] = attr.ib(None)
    monitoring: Optional[MonitoringMetadata] = attr.ib(None)

    @owners.validator
    def validate_owners(self, attribute, value):
        """Check that provided email addresses or github identities for owners are valid."""
        if not all(map(lambda e: is_email_or_github_identity(e), value)):
            raise ValueError(f"Invalid email or Github identity for owners: {value}.")

    @labels.validator
    def validate_labels(self, attribute, value):
        """Check that labels are valid."""
        for key, label in value.items():
            if key == "review_bugs" and label != "":
                if isinstance(label, list):
                    for bug in label:
                        if not Metadata.is_valid_label(str(bug)):
                            raise ValueError(f"Invalid label format: {bug}")
                else:
                    raise ValueError("Error: review_bugs needs to be a list.")
            elif not isinstance(label, bool):
                if not Metadata.is_valid_label(str(key)):
                    raise ValueError(f"Invalid label format: {key}")
                elif not Metadata.is_valid_label(str(label)) and label != "":
                    raise ValueError(f"Invalid label format: {label}")

    @staticmethod
    def is_valid_label(label):
        """
        Check if a label has the right format.

        Only hyphens (-), underscores (_), lowercase characters, and
        numbers are allowed. International characters are not allowed.

        Keys have a minimum length of 1 character and a maximum length of
        63 characters, and cannot be empty. Values can be empty, and have
        a maximum length of 63 characters.
        """
        return re.fullmatch(r"[0-9a-z-_]{1,63}", label) is not None

    @staticmethod
    def is_metadata_file(file_path):
        """
        Check if the provided file is a metadata file.

        Checks if the name and file format match the metadata file requirements.
        """
        # todo: we should probably also check if the file actually exists etc.
        return os.path.basename(file_path) == METADATA_FILE

    @classmethod
    def of_table(cls, dataset, table, version, target_dir):
        """
        Return metadata that is associated with the provided table.

        The provided directory is searched for metadata files and is expected to
        have the following structure: /<dataset>/<table>_<version>/metadata.yaml.
        """
        path = os.path.join(target_dir, dataset, table + "_" + version)
        metadata_file = os.path.join(path, METADATA_FILE)
        cls = Metadata.from_file(metadata_file)
        return cls

    @classmethod
    def from_file(cls, metadata_file):
        """Parse metadata from the provided file and create a new Metadata instance."""
        friendly_name = None
        description = None
        owners = []
        labels = {}
        scheduling = {}
        bigquery = None
        schema = None
        workgroup_access = None
        references = {}
        external_data = None
        deprecated = False
        deletion_date = None
        monitoring = None

        with open(metadata_file, "r") as yaml_stream:
            try:
                metadata = yaml.safe_load(yaml_stream)
                table_name = str(Path(metadata_file).parent.name)
                friendly_name = metadata.get(
                    "friendly_name", string.capwords(table_name.replace("_", " "))
                )
                description = metadata.get(
                    "description",
                    "Please provide a description for the query",
                )

                if "labels" in metadata:
                    for key, label in metadata["labels"].items():
                        if isinstance(label, bool):
                            # publish key-value pair with bool value as tag
                            if label:
                                labels[str(key)] = ""
                        elif isinstance(label, list):
                            labels[str(key)] = list(map(str, label))
                        else:
                            # all other pairs get published as key-value pair label
                            labels[str(key)] = str(label)

                if "scheduling" in metadata:
                    scheduling = metadata["scheduling"]
                    if "dag_name" in scheduling and cls.is_valid_label(
                        scheduling["dag_name"]
                    ):
                        labels["dag"] = scheduling["dag_name"]

                if "bigquery" in metadata and metadata["bigquery"]:
                    converter = cattrs.BaseConverter()
                    bigquery = converter.structure(
                        metadata["bigquery"], BigQueryMetadata
                    )

                if "owners" in metadata:
                    owners = metadata["owners"]
                    owner_idx = 1
                    for owner in filter(is_email, owners):
                        label = owner.split("@")[0]
                        if Metadata.is_valid_label(label):
                            labels[f"owner{owner_idx}"] = label
                            owner_idx += 1

                if "schema" in metadata:
                    converter = cattrs.BaseConverter()
                    schema = converter.structure(metadata["schema"], SchemaMetadata)

                if "workgroup_access" in metadata:
                    converter = cattrs.BaseConverter()
                    workgroup_access = converter.structure(
                        metadata["workgroup_access"], List[WorkgroupAccessMetadata]
                    )

                if "references" in metadata:
                    references = metadata["references"]

                if "external_data" in metadata:
                    converter = cattrs.BaseConverter()
                    external_data = converter.structure(
                        metadata["external_data"], ExternalDataMetadata
                    )
                if "deprecated" in metadata:
                    deprecated = metadata["deprecated"]
                if "deletion_date" in metadata:
                    deletion_date = metadata["deletion_date"]

                if "monitoring" in metadata:
                    converter = cattrs.BaseConverter()
                    monitoring = converter.structure(
                        metadata["monitoring"], MonitoringMetadata
                    )

                    if "partition_column" in metadata["monitoring"]:
                        # check if partition column metadata has been set explicitly;
                        # needed for monitoring config validation for views where partition
                        # column needs to be set explicitly
                        monitoring.partition_column_set = True

                return cls(
                    friendly_name,
                    description,
                    owners,
                    labels,
                    scheduling,
                    bigquery,
                    schema,
                    workgroup_access,
                    references,
                    external_data,
                    deprecated,
                    deletion_date,
                    monitoring,
                )
            except yaml.YAMLError as e:
                raise e

    @classmethod
    def of_query_file(cls, sql_file):
        """Return the metadata that is associated with the provided SQL file."""
        path, _ = os.path.split(sql_file)
        metadata_file = os.path.join(path, METADATA_FILE)
        cls = Metadata.from_file(metadata_file)
        return cls

    def write(self, file):
        """Write metadata information to the provided file."""
        converter = cattrs.BaseConverter()
        metadata_dict = converter.unstructure(self)

        if metadata_dict["scheduling"] == {}:
            del metadata_dict["scheduling"]

        if metadata_dict["labels"]:
            for label_key, label_value in metadata_dict["labels"].items():
                # handle tags
                if label_value == "":
                    metadata_dict["labels"][label_key] = True

        if "description" in metadata_dict:
            metadata_dict["description"] = Literal(metadata_dict["description"])

        if metadata_dict["schema"] is None:
            del metadata_dict["schema"]

        if metadata_dict["workgroup_access"] is None:
            del metadata_dict["workgroup_access"]

        if metadata_dict["external_data"] is None:
            del metadata_dict["external_data"]

        if not metadata_dict["deprecated"]:
            del metadata_dict["deprecated"]

        if not metadata_dict["deletion_date"]:
            del metadata_dict["deletion_date"]

        if not metadata_dict["monitoring"]:
            del metadata_dict["monitoring"]

        file.write_text(
            yaml.dump(
                converter.unstructure(metadata_dict),
                default_flow_style=False,
                sort_keys=False,
            )
        )

    def is_public_bigquery(self):
        """Return true if the public_bigquery flag is set."""
        return "public_bigquery" in self.labels

    def is_public_json(self):
        """Return true if the public_json flag is set."""
        return "public_json" in self.labels

    def is_incremental(self):
        """Return true if the incremental flag is set."""
        return "incremental" in self.labels

    def is_incremental_export(self):
        """Return true if the incremental_export flag is set."""
        return "incremental_export" in self.labels

    def review_bugs(self):
        """Return the bug ID of the data review bug in bugzilla."""
        return self.labels.get("review_bugs", None)

    def set_bigquery_partitioning(
        self, field, partition_type, required, expiration_days=None
    ):
        """Update the BigQuery partitioning metadata."""
        clustering = None
        if self.bigquery and self.bigquery.clustering:
            clustering = self.bigquery.clustering

        self.bigquery = BigQueryMetadata(
            time_partitioning=PartitionMetadata(
                field=field,
                type=PartitionType(partition_type),
                require_partition_filter=required,
                expiration_days=expiration_days,
            ),
            clustering=clustering,
        )

    def set_bigquery_clustering(self, clustering_fields):
        """Update the BigQuery partitioning metadata."""
        if self.bigquery:
            time_partitioning = self.bigquery.time_partitioning
            range_partitioning = self.bigquery.range_partitioning

            self.bigquery = BigQueryMetadata(
                time_partitioning=time_partitioning,
                range_partitioning=range_partitioning,
                clustering=ClusteringMetadata(fields=clustering_fields),
            )


@attr.s(auto_attribs=True)
class DatasetMetadata:
    """
    Representation of a dataset-level metadata configuration.

    Uses attrs to simplify the class definition and provide validation.
    Docs: https://www.attrs.org
    """

    friendly_name: str = attr.ib()
    description: str = attr.ib()
    dataset_base_acl: str = attr.ib()
    user_facing: bool = attr.ib(False)
    labels: Dict = attr.ib({})
    default_table_workgroup_access: Optional[List[Dict[str, Any]]] = attr.ib(None)
    default_table_expiration_ms: str = attr.ib(None)
    workgroup_access: list = attr.ib(DEFAULT_WORKGROUP_ACCESS)
    syndication: Dict = attr.ib({})

    def __attrs_post_init__(self):
        """Set default table workgroup access to workgroup access."""
        if self.default_table_workgroup_access is None:
            self.default_table_workgroup_access = self.workgroup_access

    @staticmethod
    def is_dataset_metadata_file(file_path):
        """
        Check if the provided file is a metadata file.

        Checks if the name and file format match the metadata file requirements.
        """
        # todo: we should probably also check if the file actually exists etc.
        return os.path.basename(file_path) == DATASET_METADATA_FILE

    def write(self, file):
        """Write dataset metadata information to the provided file."""
        metadata_dict = self.__dict__
        if metadata_dict["labels"]:
            for label_key, label_value in metadata_dict["labels"].items():
                # handle tags
                if label_value == "":
                    metadata_dict["labels"][label_key] = True

        if "description" in metadata_dict:
            metadata_dict["description"] = Literal(metadata_dict["description"])

        if "default_table_workgroup_access" in metadata_dict:
            metadata_dict["default_table_workgroup_access"] = metadata_dict[
                "default_table_workgroup_access"
            ]

        converter = cattrs.BaseConverter()
        file.write_text(
            yaml.dump(
                converter.unstructure(metadata_dict),
                default_flow_style=False,
                sort_keys=False,
            )
        )

    @classmethod
    def from_file(cls, metadata_file):
        """Parse dataset metadata from the provided file.

        Returns a new DatasetMetadata instance.
        """
        with open(metadata_file, "r") as yaml_stream:
            try:
                metadata = yaml.safe_load(yaml_stream)
                return cls(**metadata)
            except yaml.YAMLError as e:
                raise e



"""Utility functions for scheduling queries."""

import re
from datetime import datetime

TIMEDELTA_RE = re.compile(
    r"^-?((?P<hours>\d+)h)?((?P<minutes>\d+)m)?((?P<seconds>\d+)s)?$"
)


def is_timedelta_string(s):
    """
    Check whether a provided string is in a timedelta format.

    Timedeltas in configs are specified like: 1h, 30m, 1h15m, ...
    """
    return TIMEDELTA_RE.match(s)


def validate_timedelta_string(s):
    """Raise an error if the provided string is not in a valid timedelta format."""
    if not is_timedelta_string(s):
        raise ValueError(
            f"Invalid timedelta value '{s}'."
            " Timedeltas should be specified like '1h', '45m', '10s', '1h30m', etc."
        )


def is_date_string(s):
    """Check whether a string is a valid date string formatted like YYYY-MM-DD."""
    try:
        datetime.strptime(s, "%Y-%m-%d")
    except ValueError:
        return False

    return True


def is_email(s):
    """Check whether the provided string is a valid email address."""
    # https://stackoverflow.com/questions/8022530/how-to-check-for-valid-email-address
    return re.match(r"[^@]+@[^@]+\.[^@]+", s)


def is_github_identity(s):
    """Check if the given string matches the format of a Github identity."""
    return re.match(r"[@mozilla]+\/[a-zA-Z0-9]+", s)


def is_email_or_github_identity(s):
    """Check if the given string is either an email or a Github identity."""
    return is_email(s) or is_github_identity(s)


DAG_NAME_RE = re.compile("^(private_)?bqetl_.+$")


def is_valid_dag_name(name):
    """Check whether the DAG name is valid."""
    return DAG_NAME_RE.match(name)


# https://stackoverflow.com/questions/14203122/create-a-regular-expression-for-cron-statement
SCHEDULE_INTERVAL_RE = re.compile(
    r"^(once|hourly|daily|weekly|monthly|yearly|"
    r"((((\d+,)+\d+|(\d+(\/|-)\d+)|\d+|\*/\d+|\*) ?){5,7})|"
    r"((\d+h)?(\d+m)?(\d+s)?))$"
)


def is_schedule_interval(interval):
    """Check whether the provided string is a valid schedule interval."""
    return SCHEDULE_INTERVAL_RE.match(interval)


def schedule_interval_delta(schedule_interval1, schedule_interval2):
    """Return the time delta between two schedule intervals as timedelta string."""
    if not is_schedule_interval(schedule_interval1) or not is_schedule_interval(
        schedule_interval2
    ):
        return None

    aliases = {
        "yearly": "0 0 1 1 *",
        "monthly": "0 0 1 * *",
        "weekly": "0 0 * * 0",
        "daily": "0 0 * * *",
        "hourly": "0 * * * *",
        "once": "* * * * *",
    }

    cron_regex = re.compile(
        r"^(?P<minutes>\d+) (?P<hours>\d+) (?P<day>\d+) (?P<month>\d+) (?P<dow>\d+)$"
    )

    if schedule_interval1 in aliases:
        schedule_interval1 = aliases[schedule_interval1]

    if schedule_interval2 in aliases:
        schedule_interval2 = aliases[schedule_interval2]

    si1 = schedule_interval1.replace("*", "0")
    si2 = schedule_interval2.replace("*", "0")

    if cron_regex.match(si2) is None or cron_regex.match(si1) is None:
        return None

    parts1 = cron_regex.match(si1).groupdict()
    parts2 = cron_regex.match(si2).groupdict()
    # delta in seconds
    delta = 0
    delta += (int(parts2["hours"]) - int(parts1["hours"])) * 60 * 60
    delta += (int(parts2["minutes"]) - int(parts1["minutes"])) * 60
    delta += (int(parts2["day"]) - int(parts1["day"])) * 24 * 60 * 60

    # todo handle month and day of week

    return f"{delta}s"



"""PyTest plugin for collecting flake8 tests on python scripts."""

from pytest_flake8 import Flake8Item

from . import is_python_executable


# adapted from
# https://github.com/tholo/pytest-flake8/blob/1.0.6/pytest_flake8.py#L59-L82
def pytest_collect_file(parent, path):
    """Collect flake8 tests."""
    config = parent.config
    if config.option.flake8 and is_python_executable(path):
        flake8ignore = config._flake8ignore(path)
        if flake8ignore is not None:
            item = Flake8Item.from_parent(parent, fspath=path)
            item.flake8ignore = flake8ignore
            item.maxlength = config._flake8maxlen
            item.maxcomplexity = config._flake8maxcomplexity
            item.showshource = config._flake8showshource
            item.statistics = config._flake8statistics
            return item



from pathlib import Path

import click
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.cli.utils import use_cloud_function_option
from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.schema import Schema
from bigquery_etl.util.common import write_sql

CRASH_TABLES = [
    ("moz-fx-data-shared-prod", "firefox_desktop_stable", "crash_v1"),
    ("moz-fx-data-shared-prod", "firefox_crashreporter_stable", "crash_v1"),
]


@click.command("generate")
@click.option(
    "--target-project",
    "--target_project",
    help="Which project the queries should be written to.",
    default="moz-fx-data-shared-prod",
)
@click.option(
    "--output-dir",
    "--output_dir",
    help="The location to write to. Defaults to sql/.",
    default=Path("sql"),
    type=click.Path(file_okay=False),
)
@use_cloud_function_option
def generate(target_project, output_dir, use_cloud_function):
    schemas = {
        f"{project}.{dataset}.{table}": Schema.for_table(
            project=project,
            dataset=dataset,
            table=table,
            partitioned_by="submission_timestamp",
            use_cloud_function=use_cloud_function,
        )
        for project, dataset, table in CRASH_TABLES
    }

    combined_schema = Schema.empty()
    for table_name, schema in schemas.items():
        if len(schema.schema["fields"]) == 0:
            raise ValueError(
                f"Could not get schema for {table_name} from dry run, "
                f"possible authentication issue."
            )
        combined_schema.merge(schema)

    fields_per_table = {
        table_name: schema.generate_compatible_select_expression(
            combined_schema,
            unnest_structs=False,
        )
        for table_name, schema in schemas.items()
    }

    template_dir = Path(__file__).parent / "templates"
    env = Environment(loader=FileSystemLoader(template_dir))

    template = env.get_template("desktop_crashes.view.sql")
    query = template.render(tables=fields_per_table, project_id=target_project)

    write_sql(
        Path(output_dir) / target_project,
        f"{target_project}.firefox_desktop.desktop_crashes",
        "view.sql",
        reformat(query),
    )


if __name__ == "__main__":
    generate()



"""clients_daily_histogram_aggregates query generator."""

import argparse
import sys
from typing import Dict, List

from jinja2 import Environment, PackageLoader

from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.util.probe_filters import get_etl_excluded_probes_quickfix

from .utils import get_schema, ping_type_from_table

ATTRIBUTES = ",".join(
    [
        "sample_id",
        "client_id",
        "ping_type",
        "submission_date",
        "os",
        "app_version",
        "app_build_id",
        "channel",
    ]
)


def render_main(**kwargs):
    """Create a SQL query for the clients_daily_histogram_aggregates dataset."""
    env = Environment(loader=PackageLoader("bigquery_etl", "glam/templates"))
    main_sql = env.get_template("clients_daily_histogram_aggregates_v1.sql")
    return reformat(main_sql.render(**kwargs))


def get_distribution_metrics(schema: Dict) -> Dict[str, List[str]]:
    """Find all distribution-like metrics in a Glean table.

    Metric types are defined in the Glean documentation found here:
    https://mozilla.github.io/glean/book/user/metrics/index.html
    """
    metric_type_set = {
        "timing_distribution",
        "memory_distribution",
        "custom_distribution",
        "labeled_timing_distribution",
        "labeled_custom_distribution",
    }
    metrics: Dict[str, List[str]] = {metric_type: [] for metric_type in metric_type_set}
    excluded_metrics = get_etl_excluded_probes_quickfix("fenix")

    # Iterate over every element in the schema under the metrics section and
    # collect a list of metric names.
    for root_field in schema:
        if root_field["name"] != "metrics":
            continue
        for metric_field in root_field["fields"]:
            metric_type = metric_field["name"]
            if metric_type not in metric_type_set:
                continue
            for field in metric_field["fields"]:
                if field["name"] not in excluded_metrics:
                    metrics[metric_type].append(field["name"])
    return metrics


def get_metrics_sql(metrics: Dict[str, List[str]]) -> dict[str, str]:
    """Return SQL snippets to fetch metrics' data. Split by labeled and unlabeled."""
    items = []
    for metric_type, metric_names in metrics.items():
        for name in metric_names:
            path = f"metrics.{metric_type}.{name}"
            value_path = f"{path}.values"
            items.append((name, metric_type, value_path))

    # build the snippets for labeled and unlabeled metrics
    labeled = []
    unlabeled = []
    for name, metric_type, value_path in sorted(items):
        if metric_type.startswith("labeled"):
            labeled.append(
                f"""
                (
                    "{name}",
                    "{metric_type}",
                    ARRAY
                    (
                        SELECT(key, value.values)
                        FROM UNNEST(metrics.{metric_type}.{name})
                    )
                )
                """
            )
        else:
            unlabeled.append(f"""('', "{name}", "{metric_type}", {value_path})""")

    return {
        "labeled": ",".join(labeled).strip(),
        "unlabeled": ",".join(unlabeled).strip(),
    }


def main():
    """Print a clients_daily_scalar_aggregates query to stdout."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--no-parameterize",
        action="store_true",
        help="Generate a query without parameters",
    )
    parser.add_argument(
        "--source-table",
        type=str,
        help="Name of Glean table",
        default="org_mozilla_fenix_stable.metrics_v1",
    )
    parser.add_argument(
        "--product",
        type=str,
        default="org_mozilla_fenix",
    )
    args = parser.parse_args()

    # If set to 1 day, then runs of copy_deduplicate may not be done yet
    submission_date = (
        "date_sub(current_date, interval 2 day)"
        if args.no_parameterize
        else "@submission_date"
    )
    header = (
        "-- Query generated by: "
        "python3 -m bigquery_etl.glam.clients_daily_histogram_aggregates "
        f"--source-table {args.source_table}"
        + (" --no-parameterize" if args.no_parameterize else "")
    )

    schema = get_schema(args.source_table)
    distributions = get_distribution_metrics(schema)
    metrics_sql = get_metrics_sql(distributions)
    if not metrics_sql["labeled"] or not metrics_sql["unlabeled"]:
        print(header)
        print("-- Empty query: no probes found!")
        sys.exit(1)
    print(
        render_main(
            header=header,
            source_table=args.source_table,
            submission_date=submission_date,
            attributes=ATTRIBUTES,
            histograms=metrics_sql["unlabeled"],
            labeled_histograms=metrics_sql["labeled"],
            ping_type=ping_type_from_table(args.source_table),
        )
    )


if __name__ == "__main__":
    main()



"""bigquery-etl CLI generate command."""

import importlib.util
import sys
from inspect import getmembers
from pathlib import Path

import rich_click as click

from bigquery_etl.cli.utils import is_valid_project, use_cloud_function_option
from bigquery_etl.config import ConfigLoader

SQL_GENERATORS_DIR = "sql_generators"
GENERATE_COMMAND = "generate"
ROOT = Path(__file__).parent.parent.parent


def generate_group(sql_generators_dir):
    """Create the CLI group for the generate command."""
    commands = []

    # import sql_generators module
    spec = importlib.util.spec_from_file_location(
        sql_generators_dir.name, (sql_generators_dir / "__init__.py").absolute()
    )
    module = importlib.util.module_from_spec(spec)
    sys.modules["sql_generators"] = module

    for path in sql_generators_dir.iterdir():
        if "__pycache__" in path.parts:
            # Ignore pycache subdirectories
            continue
        if path.is_dir() and (path / "__init__.py").exists():
            # get Python modules for generators
            spec = importlib.util.spec_from_file_location(
                path.name, (path / "__init__.py").absolute()
            )
            # import and execute the module so that we can access
            # methods that are defined in the module
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            # find the `generate` click command in the module by
            # iterating through all the members of the module
            members = getmembers(module)
            generate_cmd = [cmd[1] for cmd in members if cmd[0] == GENERATE_COMMAND]
            if len(generate_cmd) > 0:
                if isinstance(generate_cmd[0], click.Command):
                    # rename command to name of query generator
                    cmd = generate_cmd[0]
                    cmd.name = path.name
                    commands.append(cmd)

    # add commands for generating queries to `generate` click group
    return click.Group(
        name="generate", commands=commands, help="Commands for generating SQL queries."
    )


# expose click command group
generate = generate_group(
    Path(
        ConfigLoader.get(
            "default",
            "sql_generators_dir",
            fallback=ConfigLoader.get(
                "default", "sql_generators_dir", fallback=ROOT / SQL_GENERATORS_DIR
            ),
        )
    )
)


@generate.command(help="Run all query generators", name="all")
@click.option(
    "--output-dir",
    "--output_dir",
    help="Output directory generated SQL is written to",
    type=click.Path(file_okay=False),
    default="sql",
)
@click.option(
    "--target-project",
    "--target_project",
    help="GCP project ID",
    default=ConfigLoader.get("default", "project", fallback="moz-fx-data-shared-prod"),
    callback=is_valid_project,
)
@click.option(
    "--ignore",
    "-i",
    help="Do not run the listed SQL generators",
    default=[],
    multiple=True,
)
@use_cloud_function_option
@click.pass_context
def generate_all(ctx, output_dir, target_project, ignore, use_cloud_function):
    """Run all SQL generators."""
    click.echo(f"Generating SQL content in {output_dir}.")
    click.echo(ROOT / SQL_GENERATORS_DIR)

    def generator_command_sort_key(command):
        match command.name:
            # Run `glean_usage` after `stable_views` because both update `dataset_metadata.yaml` files
            # and we want the `glean_usage` updates to take precedence.
            case "stable_views":
                return (1, command.name)
            case "glean_usage":
                return (2, command.name)
            # Run `derived_view_schemas` last in case other SQL generators create derived views.
            case "derived_view_schemas":
                return (4, command.name)
            case _:
                return (3, command.name)

    for cmd in sorted(generate.commands.values(), key=generator_command_sort_key):
        if cmd.name != "all" and cmd.name not in ignore:
            ctx.invoke(
                cmd,
                output_dir=output_dir,
                target_project=target_project,
                use_cloud_function=use_cloud_function,
            )



#!/usr/bin/env python3

"""Put data from region-us.INFORMATION_SCHEMA.JOBS_BY_ORGANIZATION into a table."""

from argparse import ArgumentParser
from datetime import date, timedelta

from google.cloud import bigquery

DEFAULT_PROJECTS = [
    "mozdata",
    "moz-fx-data-shared-prod",
    "moz-fx-data-marketing-prod",
    "moz-fx-data-bq-data-science",
]

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--date", required=True, type=date.fromisoformat
)  # expect string with format yyyy-mm-dd
parser.add_argument("--project", default="moz-fx-data-shared-prod")
parser.add_argument("--destination_project", default="moz-fx-data-shared-prod")
parser.add_argument("--destination_dataset", default="monitoring_derived")
parser.add_argument("--destination_table", default="jobs_by_organization_v1")


def create_query(job_date: date, project: str):
    """Create query with filter for source projects."""
    return f"""
        SELECT
          cache_hit,
          creation_time,
          destination_table,
          end_time,
          error_result,
          job_id,
          job_type,
          parent_job_id,
          priority,
          project_id,
          project_number,
          referenced_tables,
          reservation_id,
          start_time,
          state,
          statement_type,
          timeline,
          total_bytes_billed,
          total_bytes_processed,
          total_modified_partitions,
          total_slot_ms,
          user_email,
          query_info.resource_warning as query_info_resource_warning,
          query_info.query_hashes.normalized_literals as query_info_query_hashes_normalized_literals,
          transferred_bytes,
          DATE(creation_time) as creation_date
        FROM
          `{project}.region-us.INFORMATION_SCHEMA.JOBS_BY_ORGANIZATION`
        WHERE
          DATE(creation_time) = '{job_date.isoformat()}'
          AND (
                (
                  project_id IN UNNEST({DEFAULT_PROJECTS})
                  )
                OR EXISTS
                (
                  SELECT * FROM UNNEST(referenced_tables) AS ref_tables
                    WHERE ref_tables.project_id IN UNNEST({DEFAULT_PROJECTS})
                )
                OR
                (
                    destination_table.project_id IN UNNEST({DEFAULT_PROJECTS})
                )
            )
    """


def main():
    """Run query."""
    args = parser.parse_args()
    project = args.project

    # reprocess previous date to update jobs completed after the last run
    for partition_date in (args.date - timedelta(days=1), args.date):
        partition = partition_date.strftime("%Y%m%d")
        destination_table = (
            f"{args.destination_project}.{args.destination_dataset}"
            f".{args.destination_table}${partition}"
        )

        client = bigquery.Client(project)

        query = create_query(partition_date, project)
        job_config = bigquery.QueryJobConfig(
            destination=destination_table,
            write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
        )
        result = client.query(query, job_config=job_config).result()

        print(f"Wrote {result.total_rows} rows to {destination_table}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Determine stable table average ping sizes."""

from argparse import ArgumentParser
from fnmatch import fnmatchcase
from functools import partial
from multiprocessing.pool import ThreadPool

from google.cloud import bigquery

parser = ArgumentParser(description=__doc__)
parser.add_argument("--date", required=True)  # expect string with format yyyy-mm-dd
parser.add_argument("--project", default="moz-fx-data-shared-prod")
parser.add_argument("--dataset", default="*_stable")  # pattern
parser.add_argument("--destination_dataset", default="monitoring_derived")
parser.add_argument("--destination_table", default="average_ping_sizes_v1")


def get_tables(client, project, dataset):
    """Returns list of all available tables."""
    sql = f"SELECT dataset_id, table_id FROM `{project}.{dataset}.__TABLES__`"

    try:
        result = client.query(sql).result()
        return [(row.dataset_id, row.table_id) for row in result]
    except Exception as e:
        print(f"Error querying dataset {dataset}: {e}")

    return []


def get_average_ping_size_json(client, date, table):
    """Returns the average ping sizes for a specific date."""
    print(table)
    try:
        dataset_id = table[0]
        table_id = table[1]

        sql = f"""
            WITH total_pings AS (
                SELECT COUNT(*) as total
                FROM {dataset_id}.{table_id}
                WHERE DATE(submission_timestamp) = '{date}'
            ) SELECT
                SAFE_DIVIDE(byte_size, total) AS average_ping_size,
                byte_size as total_byte_size,
                total as row_count
            FROM total_pings, `moz-fx-data-shared-prod.monitoring.stable_and_derived_table_sizes`
            WHERE submission_date = '{date}' AND dataset_id = '{dataset_id}' AND table_id = '{table_id}'
        """

        result = list(client.query(sql).result())[0]

        return {
            "submission_date": date,
            "dataset_id": dataset_id,
            "table_id": table_id,
            "average_byte_size": result.average_ping_size,
            "total_byte_size": result.total_byte_size,
            "row_count": result.row_count,
        }
    except Exception as e:
        print(e)
        return None


def save_average_ping_sizes(
    client, average_ping_sizes, date, destination_dataset, destination_table
):
    """Writes average pings sizes for tables for a specific day to BigQuery."""
    job_config = bigquery.LoadJobConfig()
    job_config.schema = (
        bigquery.SchemaField("submission_date", "DATE"),
        bigquery.SchemaField("dataset_id", "STRING"),
        bigquery.SchemaField("table_id", "STRING"),
        bigquery.SchemaField("average_byte_size", "FLOAT64"),
        bigquery.SchemaField("total_byte_size", "INT64"),
        bigquery.SchemaField("row_count", "INT64"),
    )
    job_config.write_disposition = bigquery.job.WriteDisposition.WRITE_TRUNCATE

    partition_date = date.replace("-", "")
    client.load_table_from_json(
        average_ping_sizes,
        f"{destination_dataset}.{destination_table}${partition_date}",
        job_config=job_config,
    ).result()


def main():
    args = parser.parse_args()
    client = bigquery.Client(args.project)

    stable_datasets = [
        dataset.dataset_id
        for dataset in list(client.list_datasets())
        if fnmatchcase(dataset.dataset_id, args.dataset)
    ]

    with ThreadPool(20) as p:
        stable_tables = p.map(
            partial(get_tables, client, args.project),
            stable_datasets,
            chunksize=1,
        )

        stable_tables = [table for tables in stable_tables for table in tables]

        average_ping_sizes = p.map(
            partial(get_average_ping_size_json, client, args.date),
            stable_tables,
            chunksize=1,
        )

        average_ping_sizes = [s for s in average_ping_sizes if s is not None]
        total_row_count = sum([ping_size["row_count"] for ping_size in average_ping_sizes])

        if total_row_count == 0:
            raise RuntimeError("No tables with rows found; job may be incorrectly configured.")

        save_average_ping_sizes(
            client,
            average_ping_sizes,
            args.date,
            args.destination_dataset,
            args.destination_table,
        )


if __name__ == "__main__":
    main()



import os
from pathlib import Path

import pytest

from bigquery_etl.docs.mozfun.generate_mozfun_docs import load_with_examples

TEST_DIR = Path(__file__).parent.parent


class TestGenerateDocs:
    def test_load_with_examples_udf(self):
        input = (
            TEST_DIR
            / "data"
            / "test_docs"
            / "generated_docs"
            / "test_dataset1"
            / "udf1"
            / "README.md"
        )
        result = load_with_examples(str(input)).strip()

        assert result == "```sql\nSELECT\n  *\nFROM\n  test\n```"

    def test_load_with_examples_dataset(self):
        input = (
            TEST_DIR
            / "data"
            / "test_docs"
            / "generated_docs"
            / "test_dataset1"
            / "README.md"
        )
        result = load_with_examples(str(input)).strip()

        assert result == "# test_dataset1\n\n```sql\nSELECT\n  *\nFROM\n  test\n```"

    def test_load_with_missing_example(self, tmp_path):
        file_path = tmp_path / "ds" / "udf"
        os.makedirs(file_path)
        file = file_path / "README.md"
        file.write_text("@sql(examples/non_existing.sql)")

        with pytest.raises(FileNotFoundError):
            load_with_examples(str(input)).strip()



import os
from pathlib import Path

from bigquery_etl.routine import parse_routine

TEST_DIR = Path(__file__).parent.parent


class TestParseRoutine:
    udf_dir = TEST_DIR / "data" / "test_sql" / "moz-fx-data-test-project" / "udf"

    def test_routine_instantiation(self):
        raw_routine = parse_routine.RawRoutine(
            self.udf_dir / "test_js_udf" / "udf.sql",
            "udf.test_js_udf",
            "udf",
            "moz-fx-data-test-project",
        )

        assert raw_routine.filepath == self.udf_dir / "test_js_udf" / "udf.sql"
        assert raw_routine.name == "udf.test_js_udf"
        assert raw_routine.dataset == "udf"
        assert raw_routine.project == "moz-fx-data-test-project"
        assert raw_routine.definitions == []
        assert raw_routine.tests == []
        assert raw_routine.dependencies == []
        assert raw_routine.description == "Some description"
        assert raw_routine.is_stored_procedure is False

    def test_raw_routine_from_file(self):
        result = parse_routine.RawRoutine.from_file(
            self.udf_dir / "test_bitmask_lowest_28" / "udf.sql"
        )
        assert result.name == "udf.test_bitmask_lowest_28"
        assert result.dataset == "udf"
        assert len(result.definitions) == 1
        assert (
            result.definitions[0]
            == """CREATE OR REPLACE FUNCTION udf.test_bitmask_lowest_28() AS (
  0x0FFFFFFF
);"""
        )
        assert result.tests == []
        assert result.dependencies == []

        result = parse_routine.RawRoutine.from_file(
            self.udf_dir / "test_shift_28_bits_one_day" / "udf.sql"
        )
        assert len(result.tests) == 1
        assert (
            "assert_equals(0, udf.test_shift_28_bits_one_day(1 << 27));"
            in result.tests[0]
        )
        assert result.dependencies == ["udf.test_bitmask_lowest_28"]

    def test_parse_routine(self):
        raw_routine = parse_routine.RawRoutine.from_file(
            self.udf_dir / "test_shift_28_bits_one_day" / "udf.sql"
        )
        result = parse_routine.ParsedRoutine.from_raw(raw_routine, "SELECT 'test'")
        assert result.tests_full_sql == "SELECT 'test'"
        assert result.name == "udf.test_shift_28_bits_one_day"

    def test_routine_usages_in_text(self):
        text = (
            "SELECT hist.extract(123) AS a, event_analysis.get_count_sql('{}') AS b,"
            " udf.test_bitmask_lowest_28(1)"
        )
        result = parse_routine.routine_usages_in_text(text, project=self.udf_dir.parent)
        assert "hist.extract" in result
        assert "event_analysis.get_count_sql" in result
        assert "udf.test_bitmask_lowest_28" in result

        text = ""
        assert (
            parse_routine.routine_usages_in_text(text, project=self.udf_dir.parent)
            == []
        )

        text = "SELECT mozfun.unknown.udf(1) AS a, udf.test_js_udf(1) AS b"
        result = parse_routine.routine_usages_in_text(text, project=self.udf_dir.parent)
        assert result == ["udf.test_js_udf"]

    def test_read_routine_dir(self):
        raw_routines = parse_routine.read_routine_dir(self.udf_dir)
        assert "udf.test_shift_28_bits_one_day" in raw_routines
        assert "udf.test_safe_crc32_uuid" in raw_routines
        assert "udf.test_safe_sample_id" in raw_routines
        assert "udf.test_shift_28_bits_one_day"
        assert (
            raw_routines["udf.test_shift_28_bits_one_day"].name
            == "udf.test_shift_28_bits_one_day"
        )
        assert (
            type(raw_routines["udf.test_shift_28_bits_one_day"])
            == parse_routine.RawRoutine
        )

    def test_parse_routine_dirs(self):
        parsed_routines = list(parse_routine.parse_routines(self.udf_dir.parent))
        bitmask_lowest_28 = [
            u for u in parsed_routines if u.name == "udf.test_bitmask_lowest_28"
        ][0]
        assert type(bitmask_lowest_28) == parse_routine.ParsedRoutine

        shift_28_bits_one_day = [
            u for u in parsed_routines if u.name == "udf.test_shift_28_bits_one_day"
        ][0]
        assert (
            "assert_equals(0, udf_test_shift_28_bits_one_day(1 << 27));"
            in shift_28_bits_one_day.tests_full_sql[0]
        )

    def test_accumulate_dependencies(self):
        raw_routines = parse_routine.read_routine_dir(self.udf_dir)

        result = parse_routine.accumulate_dependencies(
            [], raw_routines, "udf.test_shift_28_bits_one_day"
        )
        assert "udf.test_shift_28_bits_one_day" in result
        assert "udf.test_bitmask_lowest_28" in result

        result = parse_routine.accumulate_dependencies(
            [], raw_routines, "udf.test_bitmask_lowest_28"
        )
        assert "udf.test_bitmask_lowest_28" in result

    def test_routine_usage_definitions(self):
        raw_routines = parse_routine.read_routine_dir(self.udf_dir)

        text = "SELECT udf.test_bitmask_lowest_28(0), udf.test_safe_sample_id('')"
        result = parse_routine.routine_usage_definitions(
            text, self.udf_dir.parent, raw_routines
        )
        assert len(result) == 11
        assert (
            "CREATE OR REPLACE FUNCTION udf.test_bitmask_lowest_28()"
            + " AS (\n  0x0FFFFFFF\n);"
            in result
        )
        assert (
            "CREATE OR REPLACE FUNCTION udf.test_safe_sample_id(client_id STRING) AS"
            + " (\n  MOD(udf.test_safe_crc32_uuid(CAST(client_id AS BYTES)), 100)\n);"
            in result
        )

    def test_sub_local_routines(self):
        data_dir = TEST_DIR / "data" / "test_sql" / "moz-fx-data-test-project"
        raw_routines = parse_routine.read_routine_dir(data_dir / "udf")
        raw_routines.update(parse_routine.read_routine_dir(data_dir / "procedure"))
        raw_routine = parse_routine.RawRoutine.from_file(
            data_dir / "udf" / "test_shift_28_bits_one_day" / "udf.sql"
        ).tests[0]

        assert "CREATE TEMP FUNCTION" not in raw_routine
        assert "CREATE TEMP FUNCTION udf_test_bitmask_lowest_28" not in raw_routine
        result = parse_routine.sub_local_routines(
            raw_routine, self.udf_dir.parent, raw_routines
        )
        assert "CREATE TEMP FUNCTION udf_test_shift_28_bits_one_day" in result
        assert "CREATE TEMP FUNCTION udf_test_bitmask_lowest_28" in result

        text = "SELECT udf.test_bitmask_lowest_28(23), mozfun.hist.extract('{}')"
        result = parse_routine.sub_local_routines(
            text, self.udf_dir.parent, raw_routines
        )
        assert "CREATE TEMP FUNCTION udf_test_bitmask_lowest_28" in result

        # There is no defn for hist.extract in the `raw_routines`,
        # so we expect this to be unreplaced
        assert "hist_extract" not in result
        assert "mozfun.hist.extract" in result

        text = "CALL procedure.test_procedure(23);"
        result = parse_routine.sub_local_routines(
            text, self.udf_dir.parent, raw_routines, stored_procedure_test=True
        )
        assert (
            "CREATE OR REPLACE PROCEDURE\n  _generic_dataset_.procedure_test_procedure"
            in result
        )

    def test_routine_tests_sql(self):
        raw_routines = parse_routine.read_routine_dir(self.udf_dir)
        raw_routine = parse_routine.RawRoutine.from_file(
            self.udf_dir / "test_shift_28_bits_one_day" / "udf.sql"
        )
        result = parse_routine.routine_tests_sql(
            raw_routine, raw_routines, self.udf_dir.parent
        )[0]
        assert "CREATE TEMP FUNCTION udf_test_shift_28_bits_one_day" in result
        assert "CREATE TEMP FUNCTION udf_test_bitmask_lowest_28" in result

        raw_routine = parse_routine.RawRoutine.from_file(
            self.udf_dir / "test_bitmask_lowest_28" / "udf.sql"
        )
        result = parse_routine.routine_tests_sql(
            raw_routine, raw_routines, self.udf_dir.parent
        )
        assert result == []

    def test_udf_description(self):
        raw_routine = parse_routine.RawRoutine.from_file(
            self.udf_dir / "test_shift_28_bits_one_day" / "udf.sql"
        )
        assert (
            raw_routine.description
            == "Shift input bits one day left and drop any bits beyond 28 days."
        )

    def test_procedure(self, tmp_path):
        text = (
            "CREATE OR REPLACE PROCEDURE procedure.test_procedure(out STRING) "
            "BEGIN "
            "SET out = mozfun.json.mode_last('{}'); "
            "END "
        )
        procedure_file = (
            tmp_path / "procedure" / "test_procedure" / "stored_procedure.sql"
        )
        os.makedirs(procedure_file.parent)
        procedure_file.write_text(text)
        result = parse_routine.RawRoutine.from_file(procedure_file)
        assert result.name == "procedure.test_procedure"
        assert len(result.definitions) == 1
        assert len(result.dependencies) == 1
        assert "json.mode_last" in result.dependencies
        assert result.tests == [text.strip()]

        text = (
            "CREATE OR REPLACE PROCEDURE procedure.test_procedure(out STRING) "
            "BEGIN "
            "SET out = ''; "
            "END "
        )
        procedure_file.write_text(text)
        result = parse_routine.RawRoutine.from_file(procedure_file)
        assert result.name == "procedure.test_procedure"
        assert len(result.definitions) == 1
        assert result.dependencies == []
        assert result.tests == [text.strip()]

    def test_procedure_from_file(self):
        result = parse_routine.RawRoutine.from_file(
            self.udf_dir.parent
            / "procedure"
            / "test_procedure"
            / "stored_procedure.sql"
        )
        assert result.name == "procedure.test_procedure"
        assert result.dataset == "procedure"
        assert len(result.definitions) == 1
        assert (
            result.definitions[0]
            == """CREATE OR REPLACE PROCEDURE
  procedure.test_procedure(out STRING)
BEGIN
  SET out = mozfun.json.mode_last('{}');
END;"""
        )
        assert result.tests == [result.definitions[0].strip()]
        assert result.dependencies == ["json.mode_last"]

    def test_procedure_tests_sql(self):
        raw_procedure = parse_routine.RawRoutine.from_file(
            self.udf_dir.parent / "procedure" / "append_hello" / "stored_procedure.sql"
        )

        raw_routines = parse_routine.read_routine_dir(self.udf_dir)
        raw_routines.update(
            parse_routine.read_routine_dir(self.udf_dir.parent / "procedure")
        )

        tests = parse_routine.routine_tests_sql(
            raw_procedure, raw_routines, self.udf_dir.parent
        )
        assert (
            "CREATE OR REPLACE PROCEDURE\n  _generic_dataset_.procedure_test_procedure"
            in tests[0]
        )
        assert (
            "CREATE OR REPLACE PROCEDURE\n  _generic_dataset_.procedure_append_hello"
            in tests[0]
        )

        assert (
            "CREATE OR REPLACE PROCEDURE\n  _generic_dataset_.procedure_test_procedure"
            in tests[1]
        )
        assert (
            "CREATE OR REPLACE PROCEDURE\n  _generic_dataset_.procedure_append_hello"
            in tests[1]
        )
        assert (
            "CREATE OR REPLACE FUNCTION _generic_dataset_.udf_test_shift_28_bits_one_day"
            in tests[1]
        )

    def test_get_routines_from_dir(self):
        routines = parse_routine.get_routines_from_dir(self.udf_dir.parent)
        assert len(routines) == 7

        routines = parse_routine.get_routines_from_dir("non-existing")
        assert len(routines) == 0



"""PyTest plugin for running sql tests."""

import datetime
import json
import os.path
import re
from functools import partial
from multiprocessing.pool import ThreadPool

import pytest
from google.api_core.exceptions import BadRequest
from google.cloud import bigquery

from ..routine import parse_routine
from ..util.common import render
from .sql_test import (
    TABLE_EXTENSIONS,
    Table,
    coerce_result,
    dataset,
    default_encoding,
    get_query_params,
    load,
    load_table,
    load_view,
    print_and_test,
    read,
)

expect_names = {f"expect.{ext}" for ext in ("yaml", "json", "ndjson")}


def pytest_configure(config):
    """Register a custom marker."""
    config.addinivalue_line("markers", "sql: mark sql tests.")


def pytest_collect_file(parent, path):
    """Collect non-python query tests."""
    if path.basename in expect_names:
        return SqlTest.from_parent(parent, fspath=path.parts()[-2])


class SqlTest(pytest.Item, pytest.File):
    """Test a SQL query."""

    def __init__(self, fspath, parent):
        """Initialize."""
        super().__init__(fspath, parent)
        self._nodeid += "::SQL"
        self.add_marker("sql")

    def reportinfo(self):
        """Set report title to `{dataset}.{table}:{test}`."""
        project, dataset, table, test = self.fspath.strpath.split(os.path.sep)[-4:]
        return super().reportinfo()[:2] + (f"{project}.{dataset}.{table}:{test}",)

    def repr_failure(self, excinfo):
        """Skip traceback for api error."""
        if excinfo.errisinstance(BadRequest):
            return str(excinfo.value)
        return super().repr_failure(excinfo)

    def _prunetraceback(self, excinfo):
        """Prune traceback to runtest method."""
        traceback = excinfo.traceback
        ntraceback = traceback.cut(path=__file__)
        excinfo.traceback = ntraceback.filter()

    def runtest(self):
        """Run."""
        test_name = self.fspath.basename
        query_name = self.fspath.dirpath().basename
        project_dir = (
            self.fspath.dirpath().dirpath().dirpath().dirname.replace("tests", "")
        )

        script_test = False

        # init tests write to dataset_query_test, instead of their default name
        path = self.fspath.dirname.replace("tests", "")
        if test_name == "test_init":
            query = render(
                "query.sql",
                template_folder=path,
                **{"is_init": lambda: True},
            )
        elif test_name == "test_script":
            script_test = True
            query = render("script.sql", template_folder=path)
        else:
            query = render("query.sql", template_folder=path)

        expect = load(self.fspath.strpath, "expect")

        tables = {}
        views = {}

        # generate tables for files with a supported table extension
        for resource in next(os.walk(self.fspath))[2]:
            if "." not in resource:
                continue  # tables require an extension
            table_name, extension = resource.rsplit(".", 1)
            if table_name.endswith(".schema") or table_name in (
                "expect",
                "query_params",
            ):
                continue  # not a table
            if extension in TABLE_EXTENSIONS or extension in ("yaml", "json"):
                if extension in TABLE_EXTENSIONS:
                    source_format = TABLE_EXTENSIONS[extension]
                    source_path = os.path.join(self.fspath.strpath, resource)
                else:
                    source_format = TABLE_EXTENSIONS["ndjson"]
                    source_path = (self.fspath.strpath, table_name)
                if "." in table_name:
                    # combine project and dataset name with table name
                    original, table_name = (
                        table_name,
                        table_name.replace(".", "_").replace("-", "_"),
                    )
                    original_pattern = (
                        r"`?(?<![._])\b"
                        + r"`?\.`?".join(original.split("."))
                        + r"\b(?![._])`?"
                    )
                    query = re.sub(original_pattern, table_name, query)
                else:
                    original = table_name

                # second check for tablename tweaks.
                # if the tablename ends with a date then need to replace that date with '*' for the
                # query text substitution to work.
                # e.g. see moz-fx-data-marketing-prod.65789850.ga_sessions_20230214
                # A query using that table uses moz-fx-data-marketing-prod.65789850.ga_sessions_*
                # with the date appended to allow for daily processing.
                try:
                    datetime.datetime.strptime(table_name[-8:], "%Y%m%d")
                except ValueError:
                    pass
                else:
                    generic_table_name = table_name[:-8] + "*"
                    generic_original = original[:-8] + "*"
                    query = query.replace(generic_original, generic_table_name)
                tables[table_name] = Table(table_name, source_format, source_path)
                print(f"Initialized {table_name}")
            elif extension == "sql":
                if "." in table_name:
                    # combine project and dataset name with table name
                    original, table_name = (
                        table_name,
                        table_name.replace(".", "_").replace("-", "_"),
                    )
                    query = query.replace(original, table_name)
                views[table_name] = read(self.fspath.strpath, resource)

        # rewrite all udfs as temporary
        query = parse_routine.sub_local_routines(query, project_dir)

        # if we're reading an initialization function, ensure that we're not
        # using a partition filter since we rely on `select * from {table}`
        query = query.replace(
            "require_partition_filter = TRUE", "require_partition_filter = FALSE"
        )

        dataset_id = "_".join(self.fspath.strpath.split(os.path.sep)[-3:])
        if "CIRCLE_BUILD_NUM" in os.environ:
            dataset_id += f"_{os.environ['CIRCLE_BUILD_NUM']}"

        bq = bigquery.Client()
        with dataset(bq, dataset_id) as default_dataset:
            with ThreadPool(8) as pool:
                pool.map(
                    partial(load_table, bq, default_dataset),
                    tables.values(),
                    chunksize=1,
                )
                pool.starmap(
                    partial(load_view, bq, default_dataset), views.items(), chunksize=1
                )

            # configure job
            res_table = bigquery.TableReference(default_dataset, query_name)

            if script_test:
                job_config = bigquery.QueryJobConfig(
                    default_dataset=default_dataset,
                    query_parameters=get_query_params(self.fspath.strpath),
                    use_legacy_sql=False,
                )

                bq.query(query, job_config=job_config).result()
                # Retrieve final state of table on init or script tests
                job = bq.query(
                    f"SELECT * FROM {dataset_id}.{query_name}", job_config=job_config
                )

            else:
                job_config = bigquery.QueryJobConfig(
                    default_dataset=default_dataset,
                    destination=res_table,
                    query_parameters=get_query_params(self.fspath.strpath),
                    use_legacy_sql=False,
                    write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
                )

                # run query
                job = bq.query(query, job_config=job_config)

            result = list(coerce_result(*job.result()))
            result.sort(
                key=lambda row: json.dumps(
                    row, sort_keys=True, default=default_encoding
                )
            )
            # make sure we encode dates correctly
            expect = json.loads(
                json.dumps(
                    sorted(
                        expect,
                        key=lambda row: json.dumps(
                            row, sort_keys=True, default=default_encoding
                        ),
                    ),
                    default=default_encoding,
                )
            )

            print_and_test(expect, result)
